"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Unittest for entrypoint, drain-swamp

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_cli_igor.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_cli_igor.py

Needs a config file to specify exact files to include / omit from report.
Will fail with exit code 1 even with 100% coverage

"""

import copy
import logging
import logging.config
import shlex
import subprocess
import traceback
from pathlib import Path
from unittest.mock import patch

import pytest  # noqa: F401
from click.testing import CliRunner

from drain_swamp.cli_igor import (
    current_version,
    do_cheats,
    entrypoint_name,
    main,
    seed,
    semantic_version_aware_build,
    setuptools_scm_key_value_pair,
    snippets_list,
    tag_version,
    validate_tag,
)
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)


def test_cli_main():
    """Minimally test package version is printed"""
    runner = CliRunner()
    # --version
    """
    cmd = ["--version"]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    assert "version" in result.stdout
    """

    # --help
    cmd = ["--help"]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    assert f"Command-line for {entrypoint_name}. Prints usage" in result.stdout


def test_seed(tmp_path, prep_pyproject_toml):
    """To CHANGES.rst, add seed, later to be replaced by a changelog versioned entry"""
    # prepare
    path_change_rst = Path(__file__).parent.joinpath(
        "_changelog_files",
        "CHANGES-empty.rst",
    )

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        cmd = ["--path", path_tmp_dir]

        # without prepare
        result = runner.invoke(seed, cmd)
        expected = "Provide an absolute path to a file. Got, "
        actual = result.stdout
        assert expected in actual

        # prepare
        path_empty = prep_pyproject_toml(
            path_change_rst, path_tmp_dir, rename="CHANGES.rst"
        )
        existing_text = path_empty.read_text()

        result = runner.invoke(seed, cmd)
        expected_out = "Updating "
        actual_out = result.stdout
        assert expected_out in actual_out

        actual_text = path_empty.read_text()
        assert existing_text != actual_text
        assert "Nothing yet." in actual_text


def test_build_cli(prep_pyproject_toml, tmp_path):
    """build package -- drain-swamp only. No refresh locking.

    See cmd within build backend instead
    """
    # pytest --showlocals --log-level INFO -k "test_build_cli" tests
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)

        # Passing folder is caught by click rather than NotADirectoryError
        path_passwd = path_tmp_dir.joinpath("password-secret.txt")
        path_passwd.touch()
        kind = "five element spirit or nature"
        cmd = [
            "--path",
            str(path_passwd),
            "--kind",
            shlex.split(kind),
        ]
        result = runner.invoke(semantic_version_aware_build, cmd)
        assert result.exit_code == 2
        del cmd

        # AssertionError
        # shlex.split("   ") --> [] which is interpreted as "[]" --> exit code 6, not 4
        # 1.2345 --> "1.2345" --> exit code 6, not 4
        cmd = [
            "--path",
            str(path_tmp_dir),
            "--kind",
            "tag",
        ]
        invalids = (
            "",
            # shlex.split("   "),
            # 1.2345,
        )
        for invalid in invalids:
            cmd_folder_invalid = copy.deepcopy(cmd)
            with (
                patch(
                    f"{g_app_name}.version_semantic.get_package_name", return_value=None
                ),
            ):
                result = runner.invoke(semantic_version_aware_build, cmd_folder_invalid)
                print(
                    f"test_cli_igor.py test_build AssertionError result.stdout {result.stdout}"
                )
                assert result.exit_code == 4
            del cmd_folder_invalid

        # ValueError
        kind = "five element spirit or nature?"
        cmd = [
            "--path",
            path_tmp_dir,
            "--kind",
            kind,
        ]
        result = runner.invoke(semantic_version_aware_build, cmd)
        assert result.exit_code == 5

        # Requires preparation of a pyproject.toml
        cmd = []
        cmd_folder_no_commits = [
            "--path",
            path_tmp_dir,
            "--kind",
            "0.0.1",
        ]
        with (
            patch(
                "subprocess.run",
                side_effect=subprocess.CalledProcessError(128, cmd),
            ),
        ):
            result = runner.invoke(semantic_version_aware_build, cmd_folder_no_commits)
            assert result.exit_code == 4
            assert len(result.output) != 0

        # prepare
        p_toml_file = Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        )
        prep_pyproject_toml(p_toml_file, path_tmp_dir)

        # build fail
        cmd = []
        cmd_folder_no_commits = [
            "--path",
            path_tmp_dir,
            "--kind",
            "0.0.1",
        ]
        with (
            patch(
                "subprocess.run",
                side_effect=subprocess.CalledProcessError(128, cmd),
            ),
        ):
            result = runner.invoke(semantic_version_aware_build, cmd_folder_no_commits)
            assert result.exit_code == 1
            assert len(result.output) != 0

        # build success
        with (
            patch(
                "subprocess.run",
                return_value=subprocess.CompletedProcess(
                    cmd,
                    returncode=0,
                    stdout="You just installed the best package and dependencies ... ever",
                ),
            ),
        ):
            result = runner.invoke(semantic_version_aware_build, cmd_folder_no_commits)
            assert result.exit_code == 0


testdata_validate_tag = (
    ("1!v1.0.1+g4b33a80.d20240129", "1.0.1", 0),
    ("0.1.1.candidate1dev1+g4b33a80.d20240129", "0.1.1rc1.dev1", 1),
)
ids_validate_tag = (
    "with epoch locals and prepended v",
    "malformed semantic ver str raise ValueError",
)


@pytest.mark.parametrize(
    "ver, expected_msg, expected_exit_code",
    testdata_validate_tag,
    ids=ids_validate_tag,
)
def test_validate_tag(ver, expected_msg, expected_exit_code):
    """Check validate a str. Is it a valid semantic version str?"""
    # pytest --showlocals --log-level INFO -k "test_validate_tag" tests
    runner = CliRunner()
    cmd = [ver]
    result = runner.invoke(validate_tag, cmd)
    if expected_exit_code == 0:
        actual = result.output
        assert actual.rstrip() == expected_msg
        assert result.exit_code == 0
    else:
        assert result.exit_code == 1
        # why semantic version str is malformed
        assert len(result.output) != 0


testdata_current_version = (
    (
        None,
        1,
    ),
    (
        "abcdefg",
        0,
    ),
)
ids_current_version = (
    "call failure",
    "call succeeds gets the current version str",
)


@pytest.mark.parametrize(
    "ret, expected_exit_code",
    testdata_current_version,
    ids=ids_current_version,
)
def test_current_version(ret, expected_exit_code, path_project_base):
    # demonstrate exit codes, but not that a ver str is returned on a successful call
    runner = CliRunner()

    cmd = ["--path", path_project_base()]
    with (patch(f"{g_app_name}.cli_igor.get_current_version", return_value=ret),):
        result = runner.invoke(current_version, cmd)

    assert result.exit_code == expected_exit_code


testdata_snippets_list = (
    (
        "docs",
        Path(__file__).parent.joinpath(
            "test_snip",
            "test_snip_harden_one_snip__with_id_.txt",
        ),
        0,
        ("little_shop_of_horrors_shrine_candles",),
    ),
    (
        "docs",
        Path(__file__).parent.joinpath(
            "test_snip",
            "test_snip_harden_No_snippet__Nothing_to_do_.txt",
        ),
        6,
        (),
    ),
    (
        "docs",
        Path(__file__).parent.joinpath(
            "_bad_snips",
            "nested-snips.txt",
        ),
        5,
        (),
    ),
)
ids_snippets_list = (
    "Has a snippet",
    "No snippets",
    "Nested snips --> ReplaceResult.VALIDATE_FAIL",
)


@pytest.mark.parametrize(
    "doc_relpath, path_confpy_src, expected_exit_code, keys_expected",
    testdata_snippets_list,
    ids=ids_snippets_list,
)
def test_snippets_list(
    doc_relpath,
    path_confpy_src,
    expected_exit_code,
    keys_expected,
    tmp_path,
    caplog,
):
    """drain-swamp list --> lists snippets in doc?/conf.py"""
    # pytest --showlocals --log-level INFO -k "test_snippets_list" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)

        # a folder --> ReplaceResult.VALIDATE_FAIL
        cmd = [
            "--path",
            path_tmp_dir,
        ]
        result = runner.invoke(snippets_list, cmd)
        assert result.exit_code == 3
        assert result.output.rstrip() == "Expected a doc/ or docs/ folder"

        # prepare
        #    docs/ folder
        path_dir = path_tmp_dir.joinpath(doc_relpath)
        path_dir.mkdir()

        result = runner.invoke(snippets_list, cmd)
        assert result.exit_code == 4
        assert (
            result.output.rstrip()
            == "Expected to find file, doc/conf.py or docs/conf.py"
        )

        # prepare
        #    [docs folder]/conf.py
        contents = path_confpy_src.read_text()
        path_confpy = path_dir.joinpath("conf.py")
        path_confpy.write_text(contents)
        assert path_confpy.exists() and path_confpy.is_file()

        result = runner.invoke(snippets_list, cmd)
        assert result.exit_code == expected_exit_code


testdata_cheats_exceptions = (
    (
        True,
        "current",
        3,
    ),
    (
        False,
        "dog food tastes better than this",
        4,
    ),
    (
        False,
        "current",
        0,
    ),
)
ids_cheats_exceptions = (
    "Expecting a folder",
    "--kind nonsense. Explicit version str invalid",
    "package_name ignored",
)


@pytest.mark.parametrize(
    "is_tmp, kind, exit_code_expected",
    testdata_cheats_exceptions,
    ids=ids_cheats_exceptions,
)
def test_cheats_exceptions(
    is_tmp, kind, exit_code_expected, path_project_base, tmp_path
):
    # pytest --showlocals --log-level INFO -k "test_cheats_exceptions" tests

    # prepare
    if is_tmp:
        path_f = tmp_path.joinpath("fish.txt")
        path_f.touch()
    else:
        path_cwd = path_project_base()
        path_f = path_cwd

    # act
    cmd = [
        "--path",
        path_f,
        "--kind",
        kind,
    ]
    runner = CliRunner()
    result = runner.invoke(do_cheats, cmd)

    # out = result.output

    tb = result.exc_info[2]
    traceback.print_tb(tb)
    # str_tb = traceback.format_tb(tb)
    # result_exc = result.exception
    pass

    exit_code_actual = result.exit_code
    assert exit_code_actual == exit_code_expected


def test_cheats_unusual(path_project_base, tmp_path):
    # pytest --showlocals --log-level INFO -k "test_cheats_unusual" tests
    path_cwd = path_project_base()
    runner = CliRunner()

    # current
    cmd = [
        "--path",
        str(path_cwd),
        "--kind",
        "current",
    ]
    result = runner.invoke(do_cheats, cmd)
    str_cheats = result.output

    tb = result.exc_info[2]
    traceback.print_tb(tb)
    # str_tb = traceback.format_tb(tb)
    assert result.exit_code == 0
    assert str_cheats is not None and isinstance(str_cheats, str)
    assert len(str_cheats) != 0

    # tag
    exit_code_expected = 0
    kind = "tag"
    cmd = ["--path", str(path_cwd), "--kind", kind]
    result = runner.invoke(do_cheats, cmd)
    exit_code_actual = result.exit_code
    assert exit_code_actual == exit_code_expected

    # no pyproject.toml --> AssertionError(6)
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        kind = "tag"
        cmd = ["--path", tmp_dir_path, "--kind", kind]
        result = runner.invoke(do_cheats, cmd)
        exit_code_actual = result.exit_code
        assert exit_code_actual == 6


testdata_write_version_normal = (
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        "tag",
        "0.0.3",
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        "0.0.2",
        "0.0.2",
    ),
)
ids_write_version_normal = (
    "version from version file",
    "explicit valid semantic version str",
)


@pytest.mark.parametrize(
    "path_toml_src, kind, expected",
    testdata_write_version_normal,
    ids=ids_write_version_normal,
)
def test_write_version_normal(
    path_toml_src,
    kind,
    expected,
    tmp_path,
    prep_pyproject_toml,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
):
    """Write version file typical situation"""
    # pytest --showlocals --log-level INFO -k "test_write_version_normal" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner(mix_stderr=False)
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        path_cwd = path_tmp_dir.joinpath("complete_awesome_perfect")
        path_cwd.mkdir()
        version_file_relpath = "src/complete_awesome_perfect/_version.py"
        path_version_file = path_cwd.joinpath(version_file_relpath)
        # prepare
        #    pyproject.toml
        prep_pyproject_toml(path_toml_src, path_cwd)
        #    folders and empty _version.py
        srcs = (version_file_relpath,)
        prepare_folders_files(srcs, path_cwd)

        #    version file -- Avoid fallback sane sem version "0.0.1"
        str_version_file = (
            """__version__ = version = '0.0.3'\n"""
            """__version_tuple__ = version_tuple = (0, 0, 3)\n\n"""
        )
        path_version_file.write_text(str_version_file)

        args = [
            "--path",
            str(path_cwd),
        ]

        args2 = copy.deepcopy(args)
        args2.extend(["--kind", kind])
        result = runner.invoke(setuptools_scm_key_value_pair, args2)
        logger.info(f"result.stderr: {result.stderr}")
        exit_code_actual = result.exit_code
        assert exit_code_actual == 0
        actual_contents = path_version_file.read_text()
        assert expected in actual_contents

        # confirm -- get tag version, not current version
        args_get = ("--path", str(path_cwd))
        result_ver = runner.invoke(tag_version, args_get)
        logger.info(f"exc: {result_ver.exception}")
        logger.info(f"err: {result_ver.stderr}")
        logger.info(f"out: {result_ver.stdout}")
        assert has_logging_occurred(caplog)

        assert result_ver.exit_code == 0
        # str_ver_stderr = result_ver.stderr.strip()
        str_ver_stdout = result_ver.stdout.strip()
        assert str_ver_stdout == expected


testdata_write_version_exceptions = (
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete-manage-pip-prod-unlock.pyproject_toml",
        ),
        "golf balls",
        5,
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "no_project_name.pyproject_toml",
        ),
        "0.0.4",
        4,
    ),
)
ids_write_version_exceptions = (
    "kind invalid --> ValueError --> 5",
    "pyproject.toml no project.name missing --> AssertionError --> 4",
)


@pytest.mark.parametrize(
    "path_config_src, kind, exit_code_expected",
    testdata_write_version_exceptions,
    ids=ids_write_version_exceptions,
)
def test_write_version_exceptions(
    path_config_src,
    kind,
    exit_code_expected,
    tmp_path,
    prep_pyproject_toml,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
):
    """Write version file exceptions"""
    # pytest --showlocals --log-level INFO -k "test_write_version_exceptions" tests
    runner = CliRunner(mix_stderr=False)
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        path_cwd = path_tmp_dir.joinpath("complete_awesome_perfect")
        path_cwd.mkdir()
        version_file_relpath = "src/complete_awesome_perfect/_version.py"
        path_version_file = path_cwd.joinpath(version_file_relpath)

        # prepare
        #    pyproject.toml
        prep_pyproject_toml(path_config_src, path_cwd)
        #    folders and empty _version.py
        srcs = (version_file_relpath,)
        prepare_folders_files(srcs, path_cwd)

        #    version file -- Avoid fallback sane sem version "0.0.1"
        str_version_file = (
            """__version__ = version = '0.0.3'\n"""
            """__version_tuple__ = version_tuple = (0, 0, 3)\n\n"""
        )
        path_version_file.write_text(str_version_file)

        args = [
            "--path",
            str(path_cwd),
            "--kind",
            kind,
        ]
        result_ver = runner.invoke(setuptools_scm_key_value_pair, args)
        exit_code_actual = result_ver.exit_code
        assert exit_code_actual == exit_code_expected


def test_tag_version(
    tmp_path,
    prep_pyproject_toml,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
):
    """Get semantic version from version file"""
    # pytest --showlocals --log-level INFO -k "test_tag_version" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner(mix_stderr=False)
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        path_cwd = path_tmp_dir.joinpath("complete_awesome_perfect")
        path_cwd.mkdir()
        version_file_relpath = "src/complete_awesome_perfect/_version.py"
        path_version_file = path_cwd.joinpath(version_file_relpath)
        path_fish = path_cwd / "fish.txt"

        # prepare

        #    folder and empty version file
        srcs = (version_file_relpath,)
        prepare_folders_files(srcs, path_cwd)

        #    version file
        str_version_file = (
            """__version__ = version = '0.0.3'\n"""
            """__version_tuple__ = version_tuple = (0, 0, 3)\n\n"""
        )
        path_version_file.write_text(str_version_file)

        # AssertionError(6)
        args = ("--path", str(path_fish))
        result_ver = runner.invoke(tag_version, args)
        exit_code_actual = result_ver.exit_code
        assert exit_code_actual == 6

        #    pyproject.toml
        path_pyproject_toml_src = Path(__file__).parent.joinpath(
            "_good_files",
            "complete-manage-pip-prod-unlock.pyproject_toml",
        )
        prep_pyproject_toml(path_pyproject_toml_src, path_cwd)

        result_ver = runner.invoke(tag_version, args)
        exit_code_actual = result_ver.exit_code
        assert exit_code_actual == 0

        with (
            patch(
                "drain_swamp.version_semantic._tag_version",
                return_value=None,
            ),
            patch(
                "drain_swamp.version_semantic._current_version",
                return_value="golf balls",
            ),
        ):
            result_ver = runner.invoke(tag_version, args)
            exit_code_actual = result_ver.exit_code
            assert exit_code_actual == 4

        """
        out = result_ver.stdout
        err = result_ver.stderr
        exc = result_ver.exception
        logger.info(f"out: {out}")
        logger.info(f"err: {err}")
        logger.info(f"exception: {exc}")
        tb = result_ver.exc_info[2]
        traceback.print_tb(tb)
        assert has_logging_occurred(caplog)
        """
        pass



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

Unittest for module, drain_swamp.backend_setupttools

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_backend_setuptools.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_backend_setuptools.py

"""

import logging
import logging.config
import sys
from pathlib import Path
from unittest.mock import patch

import pytest

from drain_swamp.backend_abc import (
    BackendType,
    get_optionals_pyproject_toml,
    get_required_pyproject_toml,
)
from drain_swamp.backend_setuptools import BackendSetupTools  # noqa: F401
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.exceptions import (
    BackendNotSupportedError,
    MissingRequirementsFoldersFiles,
    PyProjectTOMLParseError,
)
from drain_swamp.parser_in import TomlParser

if sys.version_info >= (3, 9):  # pragma: no cover
    from collections.abc import (
        Generator,
        Sequence,
    )
else:  # pragma: no cover
    from typing import (
        Generator,
        Sequence,
    )

testdata_load_factory_good = (
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "setuptools",
    ),
    pytest.param(
        Path(__file__).parent.joinpath("_good_files", "requires-none.pyproject_toml"),
        "setuptools",
        marks=pytest.mark.xfail(raises=MissingRequirementsFoldersFiles),
    ),
)
ids_load_factory_good = (
    "setuptools backend",
    "no package dependencies. Empty dict",
)


@pytest.mark.parametrize(
    "path_config, backend_expected",
    testdata_load_factory_good,
    ids=ids_load_factory_good,
)
def test_load_factory_good(
    path_config,
    backend_expected,
    tmp_path,
    caplog,
    has_logging_occurred,
    prepare_folders_files,
    prep_pyproject_toml,
):
    # pytest --showlocals --log-level INFO -k "test_load_factory_good" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # prepare -- create folders and files
    tp = TomlParser(path_config)
    d_pyproject_toml = tp.d_pyproject_toml

    # prepare -- optionals
    d_optionals = dict()
    get_optionals_pyproject_toml(
        d_optionals,
        d_pyproject_toml,
        tmp_path,
        is_bypass=True,
    )
    seq_prepare_these = list(d_optionals.values())
    logger.info(f"prepare optionals: {seq_prepare_these}")
    prepare_folders_files(seq_prepare_these, tmp_path)

    # prepare -- required
    t_required = get_required_pyproject_toml(
        d_pyproject_toml,
        tmp_path,
        is_bypass=True,
    )
    logger.info(f"prepare required (t_required): {t_required}")
    if (
        t_required is not None
        and isinstance(t_required, Sequence)
        and len(t_required) == 2
    ):
        seq_prepare_these = (t_required[1],)
        logger.info(f"prepare required: {seq_prepare_these}")
        prepare_folders_files(seq_prepare_these, tmp_path)

    #    required
    t_required = BackendType.get_required(
        d_pyproject_toml,
        path_config,
        required=None,
    )

    lst_files = list(tmp_path.glob("**/*.in"))
    logger.info(f"lst_files: {lst_files}")

    assert has_logging_occurred(caplog)

    #    pyproject.toml
    path_f = prep_pyproject_toml(path_config, tmp_path)

    # Confirm backend name
    inst_backend = BackendType.load_factory(path_f, parent_dir=tmp_path)
    assert inst_backend.path_config == path_f
    assert inst_backend.parent_dir != path_f
    assert issubclass(type(inst_backend), BackendType)
    assert backend_expected == inst_backend.backend

    # BackendType.parent_dir setter only accepts folder
    parent_dir_before = inst_backend.parent_dir
    assert parent_dir_before.exists() and parent_dir_before.is_dir()
    assert parent_dir_before == tmp_path
    #    not a folder. Coerses file --> folder
    inst_backend.parent_dir = tmp_path.joinpath(inst_backend.path_config.name)
    del inst_backend
    inst_backend = BackendType.load_factory(path_config, parent_dir=tmp_path)

    # BackendSetupTools.compose_dependencies_line
    suffixes = (
        ".lock",
        ".unlock",
    )
    for suffix in suffixes:
        gen_actual = inst_backend.compose_dependencies_line(suffix)
        assert isinstance(gen_actual, Generator)
        lst_required_line = list(gen_actual)
        if len(lst_required_line) == 0:
            # no required packages no dependencies .in file
            pass
        else:
            assert isinstance(lst_required_line[0], str)
            assert suffix in lst_required_line[0]

        logger.info(
            f"optionals: ({len(inst_backend.optionals)}): {inst_backend.optionals}"
        )

        gen_actual = inst_backend.compose_optional_lines(suffix)
        assert isinstance(gen_actual, Generator)
        optional_lines = list(gen_actual)
        set_lines = set(optional_lines)

        set_from_parts = set()
        for line in set_lines:
            set_from_parts.add(line)
        if len(lst_required_line) != 0:
            set_from_parts.add(lst_required_line[0])

        """raises MissingRequirementsFoldersFiles if both required or
        optionals are missing"""
        str_lines_all = inst_backend.compose(suffix)
        if len(str_lines_all) == 0:
            assert len(set_from_parts) == 0
        else:
            lines_all = str_lines_all.split("\n")
            set_from_all = set(lines_all)

            assert set_from_all == set_from_parts


testdata_load_factory_bad = (
    pytest.param(
        Path(__file__).parent.joinpath("_bad_files", "backend_only.pyproject_toml"),
        marks=pytest.mark.xfail(raises=PyProjectTOMLParseError),
    ),
    pytest.param(
        Path(__file__).parent.joinpath(
            "_good_files", "backend-unsupported.pyproject_toml"
        ),
        marks=pytest.mark.xfail(raises=BackendNotSupportedError),
    ),
)
ids_load_factory_bad = (
    "has malformed value without closing double quote",
    "unsupported or unknown backend",
)


@pytest.mark.parametrize(
    "path_config",
    testdata_load_factory_bad,
    ids=ids_load_factory_bad,
)
def test_load_factory_bad(
    path_config,
    tmp_path,
    prep_pyproject_toml,
):
    # pytest --showlocals --log-level INFO -k "test_load_factory_bad" tests
    # prepare
    prep_pyproject_toml(path_config, tmp_path)
    # act
    BackendType.load_factory(path_config, parent_dir=tmp_path)


testdata_load_factory_avert_disaster = (
    pytest.param(
        Path(__file__).parent.joinpath("_good_files", "requires-none.pyproject_toml"),
        None,
        marks=pytest.mark.xfail(raises=PyProjectTOMLParseError),
    ),
    pytest.param(
        Path(__file__).parent.joinpath("_good_files", "requires-none.pyproject_toml"),
        1.2345,
        marks=pytest.mark.xfail(raises=PyProjectTOMLParseError),
    ),
    pytest.param(
        Path(__file__).parent.joinpath("_good_files", "requires-none.pyproject_toml"),
        Path(__file__).parent.joinpath("conftest.py"),
        marks=pytest.mark.xfail(raises=PyProjectTOMLParseError),
    ),
)
ids_load_factory_avert_disaster = (
    "parent_dir None",
    "not a PurePath, got float",
    "parent_dir file not a folder",
)


@pytest.mark.parametrize(
    "path_config, parent_dir",
    testdata_load_factory_avert_disaster,
    ids=ids_load_factory_avert_disaster,
)
def test_load_factory_avert_disaster(
    path_config,
    parent_dir,
    tmp_path,
):
    # pytest --showlocals --log-level INFO -k "test_load_factory_avert_disaster" tests
    """pytest-mock exists, but why bother with additional dependencies"""
    with (
        patch(
            f"{g_app_name}.backend_setuptools.BackendSetupTools.path_config",
            return_value=tmp_path,
        ),
    ):
        """During testing normal for parent_dir option to be set to temp
        folder, so requirements files are relative to there instead of the
        pytest tests subfolder where the pyproject.toml resides

        Not settint it would be bad. And we are not setting it here. mock.patch
        it to avoid strange absolute paths to requirements files and file
        exists checks failing
        """
        BackendType.load_factory(path_config, parent_dir=parent_dir)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

config_settings dict would normally contain cli args passed thru by

.. code-block:: text

   python -m build -C--[option name]="[value]"

However setuptools has not implemented support for this. The build
process hooks are each ran in a subprocess. :code:`sys.argv` is not
modified to pass thru the config settings. As a temporary workaround,
**that doesn't add functionality**, a temporary config file is created
and an environment variable containing path to the temporary config file
is passed to :code:`python -m build`

.. code-block:: shell

   export DS_CONFIG_SETTINGS=/tmp/setuptools-build.toml
   cat <<-EOF > "$DS_CONFIG_SETTINGS"
   [project]
   name = "whatever"
   version = "99.99.99a1.dev6"

   [tool.config-settings]
   kind="0.0.1"
   set-lock="0"
   EOF
   python -m build --no-isolation

The above is a valid pyproject.toml file. Neither name nor version have
any meaning. In section tool.config-settings, all key/value pairs are
str; cli command options are always str

.. seealso::

   https://pluggy.readthedocs.io/en/latest/#a-toy-example
   https://github.com/pytest-dev/pytest/blob/main/src/_pytest/hookspec.py
   https://github.com/kedro-org/kedro/blob/main/kedro/framework/hooks/specs.py

"""

from __future__ import annotations

from typing import Any

import pluggy

from .constants import HOOK_NAMESPACE

hook_spec = pluggy.HookspecMarker(HOOK_NAMESPACE)


@hook_spec
def ds_before_version_infer(config_settings: dict[str, Any]) -> str | None:
    """Hook invoked, after gathering build environment packages and before
    any build requires packages execute their hooks

    So for example, if ``setuptools-scm`` is in build-system.requires,
    it's hooks are executed after drain-swamp hooks. This is
    counter-intuitive, a pleasent surprise, and known only thru experience

    :param config_settings:

        Dict containing cli args which should have been passed thru
        by :code:`python -m build -C--[option name]="[value]"`, but isn't

        So a workaround is used until setuptools can add support

    :type config_settings: dict[str, typing.Any]
    """


@hook_spec
def ds_on_version_infer(config_settings: dict[str, Any]) -> str | None:
    """Hook invoked, after gathering build environment packages. As the scm version

    :param config_settings:

       Dict containing cli args which should have been passed thru
       by :code:`python -m build -C--[option name]="[value]"`, but isn't

       So a workaround is used until setuptools can add support

    :type config_settings: dict[str, typing.Any]
    """


@hook_spec
def ds_after_version_infer(config_settings: dict[str, Any]) -> str | None:
    """Hook invoked, after gathering build environment packages.

    :param config_settings:

       Dict containing cli args which should have been passed thru
       by :code:`python -m build -C--[option name]="[value]"`, but isn't

       So a workaround is used until setuptools can add support

    :type config_settings: dict[str, typing.Any]
    """



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

Unittest for module, drain_swamp.backend_setupttools

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_parser_in.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_parser_in.py

"""

import sys
from pathlib import (
    Path,
    PurePath,
)

import pytest

from drain_swamp.exceptions import (
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)
from drain_swamp.parser_in import TomlParser

if sys.version_info >= (3, 9):  # pragma: no cover
    from collections.abc import Mapping
else:
    from typing import Mapping

testdata_unsupported_type = [
    (None,),
    ("Hello world!",),
    (1.1234,),
    (1,),
]
ids_unsupported_type = (
    "None",
    "str",
    "float",
    "int",
)


@pytest.mark.parametrize(
    "invalid",
    testdata_unsupported_type,
    ids=ids_unsupported_type,
)
def test_get_pyproject_toml_unsupported_type(invalid):
    # pytest --showlocals --log-level INFO -k "test_get_pyproject_toml_unsupported_type" tests
    with pytest.raises(PyProjectTOMLReadError):
        TomlParser(invalid, raise_exceptions=True)


testdata_random_folders_and_files = (
    pytest.param(
        Path("/etc"),
        marks=pytest.mark.xfail(raises=PyProjectTOMLReadError),
    ),
    pytest.param(
        Path("/etc/dog_shit_throwing_monkey.conf"),
        marks=pytest.mark.xfail(raises=PyProjectTOMLReadError),
    ),
    (Path("__init__.py")),
)
ids_random_folders_and_files = (
    "folder without a pyproject.toml in it",
    "absolute but nonexistent",
    "realitive path which shockingly resolves",
)


@pytest.mark.parametrize(
    "path_f",
    testdata_random_folders_and_files,
    ids=ids_random_folders_and_files,
)
def test_random_folders_and_files(path_f):
    # pytest --showlocals --log-level INFO -k "test_random_folders_and_files" tests
    # A Path but dodgy --> FileNotFoundError
    TomlParser(path_f, raise_exceptions=True)


testdata_unparsable = list(
    Path(__file__).parent.joinpath("_bad_files").glob("backend_only.pyproject_toml")
)


@pytest.mark.parametrize(
    "path_toml_src",
    testdata_unparsable,
    ids=[path.name.rsplit(".", 1)[0] for path in testdata_unparsable],
)
def test_get_pyproject_toml_bad(path_toml_src, tmp_path, prep_pyproject_toml):
    # pytest --showlocals --log-level INFO -k "test_get_pyproject_toml_bad" tests

    # file_as_str = path.read_text()
    # Provide a path, not the file contents
    path_f = prep_pyproject_toml(path_toml_src, tmp_path)
    invalids = (path_f,)
    for invalid in invalids:
        with pytest.raises(PyProjectTOMLParseError):
            TomlParser(invalid, raise_exceptions=True)
        tp = TomlParser(invalid, raise_exceptions=False)
        actual = tp.d_pyproject_toml
        assert actual is None


testdata_valid_pyproject_toml = list(
    Path(__file__).parent.joinpath("_good_files").glob("*.pyproject_toml")
)


@pytest.mark.parametrize(
    "path_toml_src",
    testdata_valid_pyproject_toml,
    ids=[path.name.rsplit(".", 1)[0] for path in testdata_valid_pyproject_toml],
)
def test_get_pyproject_toml_good(path_toml_src, tmp_path, prep_pyproject_toml):
    # pytest --showlocals --log-level INFO -k "test_get_pyproject_toml_good" tests
    path_f = prep_pyproject_toml(path_toml_src, tmp_path)
    valids = (path_f,)
    for valid in valids:
        # Provide a path, not the file contents
        tp = TomlParser(valid)
        actual_dict = tp.d_pyproject_toml
        assert isinstance(actual_dict, Mapping)
        path_actual = tp.path_file
        assert issubclass(type(path_actual), PurePath)


testdata_search_for_pyproject_toml = (
    Path(__file__).parent.parent,
    Path(__file__).parent,
)
ids_search_for_pyproject_toml = (
    "package base folder",
    "tests folder",
)


@pytest.mark.parametrize(
    "path_dir",
    testdata_search_for_pyproject_toml,
    ids=ids_search_for_pyproject_toml,
)
def test_search_for_pyproject_toml(path_dir):
    # pytest --showlocals --log-level INFO -k "test_get_pyproject_toml_dir" tests
    tp = TomlParser(path_dir, raise_exceptions=True)
    actual = tp.d_pyproject_toml
    assert isinstance(actual, Mapping)

    # classmethod TomlParser.resolve accepts str
    dir_path = str(path_dir)
    actual = TomlParser.resolve(dir_path)
    assert issubclass(type(actual), PurePath)


def test_pyproject_toml_search_fail(tmp_path):
    # pytest --showlocals --log-level INFO -k "test_pyproject_toml_search_fail" tests
    # there should be no pyproject.toml in the reverse path
    path_f = tmp_path.joinpath("some_file.txt")
    with pytest.raises(PyProjectTOMLReadError):
        TomlParser(path_f, raise_exceptions=True)

    invalids = (
        "Go team!",
        0.1234,
    )
    for invalid in invalids:
        tp = TomlParser(path_f, raise_exceptions=invalid)
        assert tp.d_pyproject_toml is None



import logging

from setuptools_scm._integration.setuptools import version_keyword as _version_keyword
from setuptools_scm.version import (
    get_local_node_and_date,
    guess_next_dev_version,
)

log = logging.getLogger("drain_swamp.monkey.wrap_version_keyword")


def version_keyword(dist, keyword, value):
    """Remove the need for a ``setup.py`` file

    In ``pyproject.toml``,

    .. code-block:: text

       [project.entry-points."distutils.setup_keywords"]
       use_scm_version = "drain_swamp.monkey.wrap_version_keyword:version_keyword"

    :param dist: API interface for interacting with setuptools
    :type dist: setuptools.Distribution
    :param keyword: Will be ignored
    :type keyword: str
    :param value: overrides dict normally supplied in setup.py
    :type value: bool | dict[str, typing.Any] | collections.abc.Callable[[], dict[str, typing.Any]]
    """
    # keyword = "use_scm_version"
    d_value = {
        "local_scheme": get_local_node_and_date,
        "version_scheme": guess_next_dev_version,
    }
    # In version_keyword _config.Configuration.from_file param _require_section=False
    _version_keyword(dist, keyword, d_value)



PRETEND_KEY: str
PRETEND_KEY_NAMED: str

__all__ = ("normalize_dist_name",)

def normalize_dist_name(dist_name: str) -> str: ...
def read_named_env(
    *,
    tool: str = ...,
    name: str,
    dist_name: str | None,
) -> str | None: ...
def _scm_key(dist_name: str) -> str: ...



import datetime
from pathlib import Path
from typing import (
    Any,
    ClassVar,
)

from .version_semantic import SemVersion

__all__ = ("SnipSphinxConf", "entrypoint_name")

entrypoint_name: str

class SnipSphinxConf:
    DOC_FOLDERS: ClassVar[tuple[str, str]]

    def __init__(self, path: Any | None = None) -> None: ...
    @classmethod
    def now(cls) -> datetime.datetime: ...
    @classmethod
    def now_to_str(cls, strftime_str: str) -> str: ...
    @property
    def path_abs(self) -> Path: ...
    def path_abs_init(self) -> None: ...
    @property
    def path_cwd(self) -> Path: ...
    @path_cwd.setter
    def path_cwd(self, val: Any | None) -> None: ...
    @property
    def SV(self) -> SemVersion | None: ...
    @property
    def author_name_left(self) -> str | None: ...
    def contents(
        self,
        kind: str,
        package_name: str,
        copyright_start_year: str,
    ) -> str: ...
    def replace(
        self,
        snippet_co: str | None = None,
    ) -> bool: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

pipenv-unlock package entrypoint

Commands:

- refresh
- is_lock
- unlock
- lock

Has pep366 support, without installing the package, can call the
source code, as long has has required dependencies installed

"""

import sys
import traceback
from pathlib import (
    Path,
    PurePath,
)

import click

# pep366 ...
# https://stackoverflow.com/a/34155199
if __name__ == "__main__" and __spec__ is None:  # pragma: no cover
    # Package not installed
    # python src/drain_swamp/cli_unlock.py unlock --snip=little_shop_of_horrors_shrine_candles
    import importlib.util

    path_d = Path(__file__).parent
    rev_mods = []
    while path_d.joinpath("__init__.py").exists():
        name = path_d.name
        path_prev = path_d
        path_d = path_d.parent
        rev_mods.append(name)
    # One level above top package --> sys.path.insert
    sys.path.insert(1, str(path_d))
    # parent (aka package) dotted path
    dotted_path = ".".join(reversed(rev_mods))

    # print(f"str(path_d): {str(path_d)}", file=sys.stderr)
    # print(f"dotted_path: {dotted_path}", file=sys.stderr)
    # print(f"path_prev: {path_prev}", file=sys.stderr)
    pass

    # import top package level
    path_top = path_prev.joinpath("__init__.py")
    spec_top = importlib.util.spec_from_file_location(name, path_top)
    mod_top = importlib.util.module_from_spec(spec_top)
    sys.modules[dotted_path] = mod_top
    spec_top.loader.exec_module(mod_top)

    # __spec__ is None. Set __spec__ rather than :code:`__package__ = dotted_path`
    dotted_path_this = f"{dotted_path}.{Path(__file__).stem}"
    spec_this = importlib.util.spec_from_file_location(dotted_path_this, Path(__file__))
    __spec__ = spec_this
elif (
    __name__ == "__main__" and isinstance(__package__, str) and len(__package__) == 0
):  # pragma: no cover
    # When package is not installed
    # python -m drain_swamp.cli_unlock lock --snip=little_shop_of_horrors_shrine_candles
    # python -m drain_swamp.cli_unlock unlock --snip=little_shop_of_horrors_shrine_candles
    dotted_path = "drain_swamp"
    path_pkg_base_dir = Path(__file__).parent.parent
    sys.path.insert(1, str(path_pkg_base_dir))

    mod = __import__(dotted_path)
    sys.modules[dotted_path] = mod

    __package__ = dotted_path
else:
    # pipenv-unlock unlock --snip=little_shop_of_horrors_shrine_candles
    # pipenv-unlock lock --snip=little_shop_of_horrors_shrine_candles
    # __package__ = "drain_swamp"
    pass

# pep366 ...done

from .backend_abc import BackendType
from .backend_setuptools import BackendSetupTools  # noqa: F401
from .constants import (
    SUFFIX_LOCKED,
    SUFFIX_UNLOCKED,
)
from .exceptions import (
    BackendNotSupportedError,
    MissingRequirementsFoldersFiles,
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)
from .lock_toggle import (
    lock_compile,
    refresh_links,
    unlock_compile,
)
from .snip import (
    ReplaceResult,
    Snip,
)

# taken from pyproject.toml
entrypoint_name = "pipenv-unlock"  # noqa: F401

help_path = "The root directory [default: pyproject.toml directory]"
help_required = "relative to --path, required dependencies .in file"
help_optionals = (
    "relative to --path, optional dependencies .in file. Can be used multiple times"
)
help_additional_folder = (
    "Additional folder(s), not already known implicitly, containing .in "
    "files. A relative_path. Can be used multiple times"
)
help_snippet_co = (
    "Snippet code, within a file, unique id of an editable region, aka snippet. "
    "Only necessary if allows for multiple snippets"
)
help_set_lock = """Force dependency lock state. "1", "true", "t", "yes", "y", or "on" to lock; "0", "false", "f", "no", "n", "off" to unlock. refreshes .lnk symlinks"""

EPILOG_LOCK_UNLOCK = """
EXIT CODES

0 -- Evidently sufficient effort put into unittesting. Job well done, beer on me!

1 -- Unused. Reason: too generic

2 -- Unused. Reason: Bad taste left in mouth after experience with argparse

3 -- path given for config file either not a file or not read write

4 -- pyproject.toml config file parse issue. Use validate-pyproject on it then try again

5 -- Backend not supported. Need to add support for that backend. Submit an issue

6 -- The pyproject.toml depends on the requirements folders and files. Create them

7 -- For locking dependencies, pip-tools package must be installed. Not installed

8 -- The snippet is invalid. Either nested snippets or start stop token out of order. Fix the snippet then try again

9 -- In pyproject.toml, there is no snippet with that snippet code
"""

EPILOG_IS_LOCK = """
EXIT CODES

0 -- is locked

1 -- is unlocked

2 -- Path not a file

3 -- file is not ok for whatever reason

4 -- Static dependencies. No tool.setuptools.dynamic section

"""

EPILOG_REFRESH = """
0 -- symlinks refreshed

2 -- An arg is invalid

3 -- path given for config file either not a file or not read write

4 -- pyproject.toml config file parse issue. Use validate-pyproject on it then try again

5 -- Backend not supported. Need to add support for that backend. Submit an issue

6 -- The pyproject.toml depends on the requirements folders and files. Create them

7 -- In pyproject.toml no section, tool.setuptools.dynamic
"""


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
)
def main():
    """Command-line for pipenv-unlock. Prints usage"""


@main.command(
    "is_lock",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_IS_LOCK,
)
@click.option(
    "-p",
    "--path",
    default=Path.cwd(),
    type=click.Path(exists=True, file_okay=True, dir_okay=True, resolve_path=True),
    help=help_path,
)
def state_is_lock(path):
    """Check dependency locked: 0 locked, 1 unlocked else error code

    If only one snippet, snippet_co is optional

    Usage

    .. code-block:: shell

       pipenv-unlock is_lock

    \f
    :param path:

       The root directory [default: pyproject.toml directory]

    :type path: pathlib.Path
    """

    # click.secho(f"path (before): {path} type {type(path)}", fg="green")
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass
    # click.secho(f"path (after): {str(path)}", fg="green")
    pass

    try:
        out = BackendType.is_locked(path)
    except (PyProjectTOMLParseError, PyProjectTOMLReadError) as e:
        msg_exc = str(e)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(3)
    except AssertionError as e:
        # Static dependencies. No tool.setuptools.dynamic section
        msg_exc = str(e)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(4)

    if out:
        sys.exit(0)
    else:
        sys.exit(1)


@main.command(
    "lock",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_LOCK_UNLOCK,
)
@click.option(
    "-p",
    "--path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-r",
    "--required",
    default=None,
    type=(str, click.Path(exists=False, file_okay=True, dir_okay=False)),
    help=help_required,
    nargs=2,
)
@click.option(
    "-o",
    "--optional",
    "optionals",
    type=(str, click.Path(exists=False, file_okay=True, dir_okay=False)),
    help=help_optionals,
    multiple=True,
    nargs=2,
)
@click.option(
    "-d",
    "--dir",
    "additional_folders",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    multiple=True,
    help=help_additional_folder,
)
@click.option(
    "-s",
    "--snip",
    "snippet_co",
    default=None,
    type=click.STRING,
    help=help_snippet_co,
)
def dependencies_lock(path, required, optionals, additional_folders, snippet_co):
    """Lock dependencies creates (``*.lock``) files

    Symlinks (``*.lnk``) files created during build time

    Disadvantages

    1. FOSS is ``as-is``, largely unpaid work, often lacks necessary
       skillset, often doesn't care to do tedious tasks, is pressed for
       time, and live happens. These are the people supposed to be
       making packages for production use?! Having such expectations
       is ridiculous and conflicts with the human condition

    2. package quickly becomes unusable when, not if, the author is no longer
       maintaining the package

    3. Non-experts might not be using pipenv, only pip. Almost guaranteeing
       dependency hell. ``pip`` won't have what it needs to resolve
       dependency version conflicts

    4. ``pipenv`` says don't automate updating dependency lock files thru CI/CD

    5. ``pip-tools``, mistakes occur, choosing wrong dependency version.
       Had this occur. Latest version is a post release (python-dateutil-2.9.0.post0
       Sometimes choose previous release and other times not. Causing dependency hell

    Advantage

    1. Job security. Knowledgable eyeballs **must** regularly update
       dependency version locks

    2. ``pipenv`` discourages attackers setting up alternative repository hosts
       ``pypi.org`` and swapping out an obscure package with their own.

    3. The stars align in the cosmos, miraculously, all package authors regularly
       update their packages dependencys' locks. Get that warm feeling inside
       knowing we are alive, loved, and appreciated. We shout,
       ``it's a miracle!`` and be right!

    Usage

    pipenv-unlock lock --snip="little_shop_of_horrors_shrine_candles"

    or

    python src/drain_swamp/cli_unlock.py lock

    \f

    :param path:

       The root directory [default: pyproject.toml directory]

    :type path: pathlib.Path
    :param required:

       relative to --path, required dependencies .in file

    :type required: tuple[str, pathlib.Path]
    :param optionals:

       relative to --path, optional dependencies .in file. Can be used multiple times

    :type optionals: collections.abc.Sequence[tuple[str, pathlib.Path]]
    :param additional_folders:

       Additional folder(s), not already known implicitly, containing .in
       files. A relative_path. Can be used multiple times

    :type additional_folders: collections.abc.Sequence[pathlib.Path]
    :param snippet_co:

       Snippet code, within a file, unique id of an editable region, aka snippet.
       Only necessary if allows for multiple snippets

    :type snippet_co: str | None
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    # cli optionals. No user input validation yet
    # Sequence[tuple[str, pathlib.Path]] | None --> dict[str, Path]
    has_optionals = (
        optionals is not None and isinstance(optionals, tuple) and len(optionals) > 0
    )
    if has_optionals:
        d_optionals = {}
        for target, relative_path in optionals:
            if target not in d_optionals.keys():
                d_optionals[target] = relative_path
            else:  # pragma: no cover
                pass
    else:
        d_optionals = {}

    # additional folders
    #    list[str | Path] --> tuple[Path]
    s_additionals = set()
    for add_dir in additional_folders:
        if isinstance(add_dir, str):
            # Counter-intuitively, this is what click provides
            p_new = Path(add_dir)
            s_additionals.add(p_new)
        elif issubclass(type(add_dir), PurePath):  # pragma: no cover
            # This is what click should be providing
            s_additionals.add(add_dir)
        else:  # pragma: no cover
            pass
    t_add_folders = tuple(s_additionals)

    try:
        inst = BackendType.load_factory(
            path,
            required=required,
            optionals=d_optionals,
            additional_folders=t_add_folders,
        )
    except PyProjectTOMLReadError:
        msg_exc = (
            f"Either not a file or lacks read permissions. {traceback.format_exc()}"
        )
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(3)
    except PyProjectTOMLParseError:
        msg_exc = f"Cannot parse pyproject.toml. {traceback.format_exc()}"
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(4)
    except BackendNotSupportedError:
        msg_exc = f"No support yet for this Python packaging backend. {traceback.format_exc()}"
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(5)

    try:
        new_contents = inst.compose(SUFFIX_LOCKED)
    except MissingRequirementsFoldersFiles:
        msg_exc = (
            "Missing requirements folders and files. Prepare these "
            f"{traceback.format_exc()}"
        )
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(6)

    # create .lock files in their respective folders
    gen = lock_compile(inst)
    try:
        list(gen)  # execute generator
    except AssertionError:
        msg_exc = (
            "pip-tools is required to lock package dependencies. Install "
            f"it. {traceback.format_exc()}"
        )
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(7)

    # update snippet
    fname = path / "pyproject.toml"
    snip = Snip(fname, is_quiet=True)
    is_success = snip.replace(new_contents, id_=snippet_co)
    if is_success == ReplaceResult.VALIDATE_FAIL:
        msg_exc = (
            "Snippet is invalid. Validation failed. Either nested or "
            "unmatched start end tokens"
        )
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(8)
    elif is_success == ReplaceResult.NO_MATCH:
        msg_exc = (
            f"In pyproject.toml, there is no snippet with snippet code {snippet_co}"
        )
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(9)
    else:
        sys.exit(0)


@main.command(
    "unlock",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_LOCK_UNLOCK,
)
@click.option(
    "-p",
    "--path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-r",
    "--required",
    default=None,
    type=(str, click.Path(exists=False, file_okay=True, dir_okay=False)),
    help=help_required,
    nargs=2,
)
@click.option(
    "-o",
    "--optional",
    "optionals",
    type=(str, click.Path(exists=False, file_okay=True, dir_okay=False)),
    help=help_optionals,
    nargs=2,
    multiple=True,
)
@click.option(
    "-d",
    "--dir",
    "additional_folders",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    multiple=True,
    help=help_additional_folder,
)
@click.option(
    "-s",
    "--snip",
    "snippet_co",
    default=None,
    type=click.STRING,
    help=help_snippet_co,
)
def dependencies_unlock(path, required, optionals, additional_folders, snippet_co):
    """Unlock dependencies creates (``*.unlock``) files

    Symlinks (``*.lnk``) files are created during build time

    Consciously choose not to lock package dependencies

    Know that might not be available or always paying constant attention.
    Whilst package dependency lock requires frequent updating.

    It's a big ask, considering:

    - distracted by other projects and paid jobs

    - cynical

    - intend to pick up a drug habit, die, or discover girls

    Usage

    pipenv-unlock unlock --snip="little_shop_of_horrors_shrine_candles"

    or

    python src/drain_swamp/cli_unlock.py unlock

    \f

    :param path:

       The root directory [default: pyproject.toml directory]

    :type path: pathlib.Path
    :param required:

       relative to --path, required dependencies .in file

    :type required: tuple[str, pathlib.Path]
    :param optionals:

       relative to --path, optional dependencies .in file. Can be used multiple times

    :type optionals: collections.abc.Sequence[tuple[str, pathlib.Path]]
    :param additional_folders:

       Additional folder(s), not already known implicitly, containing .in
       files. A relative_path. Can be used multiple times

    :type additional_folders: collections.abc.Sequence[pathlib.Path]
    :param snippet_co:

       Snippet code, within a file, unique id of an editable region, aka snippet.
       Only necessary if allows for multiple snippets

    :type snippet_co: str | None
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    # No user input validation yet
    # Sequence[tuple[str, pathlib.Path]] | None --> dict[str, Path]
    has_optionals = (
        optionals is not None and isinstance(optionals, tuple) and len(optionals) > 0
    )
    if has_optionals:
        d_optionals = {}
        for target, relative_path in optionals:
            if target not in d_optionals.keys():
                d_optionals[target] = relative_path
            else:  # pragma: no cover
                pass
    else:
        d_optionals = {}

    # additional folders
    #    list[str | Path] --> tuple[Path]
    s_additionals = set()
    for add_dir in additional_folders:
        if isinstance(add_dir, str):
            # Counter-intuitively, this is what click provides
            p_new = Path(add_dir)
            s_additionals.add(p_new)
        elif issubclass(type(add_dir), PurePath):  # pragma: no cover
            # This is what click should be providing
            s_additionals.add(add_dir)
        else:  # pragma: no cover
            pass
    t_add_folders = tuple(s_additionals)

    try:
        inst = BackendType.load_factory(
            path,
            required=required,
            optionals=d_optionals,
            additional_folders=t_add_folders,
        )
    except PyProjectTOMLReadError:
        msg_exc = (
            f"Either not a file or lacks read permissions. {traceback.format_exc()}"
        )
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(3)
    except PyProjectTOMLParseError:
        msg_exc = f"Cannot parse pyproject.toml. {traceback.format_exc()}"
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(4)
    except BackendNotSupportedError:
        msg_exc = f"No support yet for this Python packaging backend. {traceback.format_exc()}"
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(5)

    # get contents for update to pyproject.toml
    try:
        new_contents = inst.compose(SUFFIX_UNLOCKED)
    except MissingRequirementsFoldersFiles:
        msg_exc = (
            "Missing requirements folders and files. Prepare these "
            f"{traceback.format_exc()}"
        )
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(6)

    # update snippet -- pyproject.toml
    fname = path / "pyproject.toml"
    snip = Snip(fname, is_quiet=False)
    is_success = snip.replace(new_contents, id_=snippet_co)
    if is_success == ReplaceResult.VALIDATE_FAIL:
        msg_exc = (
            "Snippet is invalid. Validation failed. Either nested or "
            "unmatched start end tokens"
        )
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(8)
    elif is_success == ReplaceResult.NO_MATCH:
        msg_exc = (
            f"In pyproject.toml, there is no snippet with snippet code {snippet_co}"
        )
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(9)
    else:
        pass

    # creates / writes .unlock files
    gen = unlock_compile(inst)
    list(gen)  # execute generator


@main.command(
    "refresh",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_REFRESH,
)
@click.option(
    "-p",
    "--path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-l",
    "--set-lock",
    "is_set_lock",
    default=None,
    type=click.BOOL,
    help=help_set_lock,
)
def create_links(path, is_set_lock):
    """From *.lock *.unlock files create symlink files *.lnk

    In ``pyproject.toml``,

    .. code-block:: text
       [build-system]
       requires = ["setuptools>=70.0.0", "wheel", "build", "setuptools_scm>=8"]
       build-backend = "setuptools.build_meta"
       backend-path = ["_req_links"]

    Executes ``_req_links.backend.py`` which calls,

    .. code-block:: shell

       python src/drain_swamp/cli_unlock.py refresh --set-lock=1

    Usage

    .. code-block:: shell

       python -m build -C--set-lock=1 --sdist

    Requires ``setuptools>=70.0.0``

    \f

    :param path:

       The root directory [default: pyproject.toml directory]

    :type path: pathlib.Path

    :param is_set_lock:

       Default None use current state. 0 make symlinks to ``.lock`` files; 1 set make
       symlinks to ``.unlock`` files.

    :type is_set_lock: int | None
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    try:
        inst = BackendType.load_factory(path)
    except PyProjectTOMLReadError:
        msg_exc = (
            f"Either not a file or lacks read permissions. {traceback.format_exc()}"
        )
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(3)
    except PyProjectTOMLParseError as e:
        # msg_exc = f"Cannot parse pyproject.toml. {traceback.format_exc()}"
        # raise click.ClickException(msg_exc)
        click.secho(str(e), fg="red", err=True)
        sys.exit(4)
    except BackendNotSupportedError as e:
        # msg_exc = f"No support yet for this Python packaging backend. {traceback.format_exc()}"
        # raise click.ClickException(msg_exc)
        click.secho(str(e), fg="red", err=True)
        sys.exit(5)

    """Create the .lnk (lock dependency state) symlinks
    This normally occurs during package build time
    PyProjectTOMLParseError and PyProjectTOMLReadError handling already done above
    """
    try:
        refresh_links(inst, is_set_lock=is_set_lock)
    except MissingRequirementsFoldersFiles:
        msg_exc = (
            "Missing requirements folders and files. Prepare these "
            f"{traceback.format_exc()}"
        )
        # raise click.ClickException(msg_exc)
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(6)
    except AssertionError:
        msg_exc = (
            "Either no pyproject.toml section, tool.setuptools.dynamic "
            "or no dependencies key"
        )
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(7)


if __name__ == "__main__":  # pragma: no cover
    main()



import logging
from pathlib import Path
from typing import (
    Any,
    ClassVar,
)

__all__ = ("ConfigSettings",)

log: logging.Logger

class ConfigSettings:
    SECTION_NAME: ClassVar[str]
    ENVIRON_KEY: ClassVar[str]
    FILE_NAME_DEFAULT: ClassVar[str]

    __slots__: ClassVar[tuple[str]] = ("_file_name",)
    def __init__(self, file_name: str | None = None) -> None: ...
    @classmethod
    def get_abs_path(cls) -> str | None: ...
    @classmethod
    def set_abs_path(cls, val: Any) -> None: ...
    @classmethod
    def remove_abs_path(cls) -> None: ...
    @property
    def file_name(self) -> str: ...
    @file_name.setter
    def file_name(self, val: Any) -> None: ...
    def read(self) -> dict[str, Any]: ...
    def write(
        self,
        path_dir: Path,
        toml_contents: str,
    ) -> None: ...
    @classmethod
    def get_section_dict(
        cls,
        path_dir: Path,
        toml_contents: str,
        file_name: str | None = None,
    ) -> dict[str, Any]: ...



"""Provides markers for declaring hook specs and implementations

.. py:data:: HOOK_NAMESPACE
   :type: str
   :value: "drain_swamp"

.. py:data:: hook_spec
   :type: pluggy.HookspecMarker

   Hook spec marker variable

.. py:data:: hook_impl
   :type: pluggy.HookimplMarker

   Hook implementation marker variable

.. seealso::

   `pluggy docs <https://pluggy.readthedocs.io/en/stable/#marking-hooks>`_
   `SOURCE <https://raw.githubusercontent.com/kedro-org/kedro/main/kedro/framework/hooks/markers.py>`_
   `LICENSE:Apache 2.0 <https://github.com/kedro-org/kedro/blob/main/LICENSE.md>`_

"""

import pluggy

from .constants import HOOK_NAMESPACE

# hook_spec = pluggy.HookspecMarker(HOOK_NAMESPACE)
hook_impl = pluggy.HookimplMarker(HOOK_NAMESPACE)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_plugin_refresh_links.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_plugin_refresh_links.py

"""

import logging
import logging.config
import shutil
from collections.abc import (
    Mapping,
    Sequence,
)
from pathlib import (
    Path,
    PurePath,
)
from unittest.mock import patch

import pytest

from drain_swamp import (
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)
from drain_swamp.backend_abc import BackendType
from drain_swamp.backend_setuptools import BackendSetupTools  # noqa: F401
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.monkey.config_settings import ConfigSettings
from drain_swamp.monkey.plugins.ds_refresh_links import (
    _is_set_lock,
    _parent_dir,
    before_version_infer,
)

testdata_is_set_lock = (
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.incorrect-section-name]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
        ),
        None,
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
        ),
        False,
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="1"\n\n"""
        ),
        True,
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
        ),
        None,
    ),
)
ids_is_set_lock = (
    "Incorrect section title. Expecting tool.config-settings",
    "set-lock provided. unlock",
    "set-lock provided. lock",
    "set-lock not provided. unlock",
)


@pytest.mark.parametrize(
    "toml_contents, is_lock_expected",
    testdata_is_set_lock,
    ids=ids_is_set_lock,
)
def test_is_set_lock(toml_contents, is_lock_expected, tmp_path):
    # pytest --showlocals --log-level INFO -k "test_is_set_lock" tests
    # prepare
    d_section = ConfigSettings.get_section_dict(tmp_path, toml_contents)

    # default is unsupported type --> unlocked
    actual = _is_set_lock(d_section, default=None)
    assert actual == is_lock_expected

    # None --> indicates to get lock state from pyproject.toml
    actual = _is_set_lock(None)
    assert actual is None

    # unsupported default --> None
    actual = _is_set_lock(None, default="")
    assert actual is None

    actual = _is_set_lock(d_section)
    assert actual == is_lock_expected

    # From cli, but not going thru :code:`python -m build`
    d_section = {"--set-lock": "1"}
    actual = _is_set_lock(d_section)
    assert actual is True


testdata_parent_dir = (
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
            """parent-dir="{}"\n"""
        )
    ),
)
ids_parent_dir = ("config_settings toml has placeholder for tmp_path",)


@pytest.mark.parametrize(
    "toml_contents",
    testdata_parent_dir,
    ids=ids_parent_dir,
)
def test_parent_dir(toml_contents, tmp_path):
    # pytest --showlocals --log-level INFO -k "test_parent_dir" tests
    str_tmp_path = str(tmp_path)

    invalids = (
        None,
        1.2,
    )
    for invalid in invalids:
        # d_section not a dict; default unsupported
        actual = _parent_dir(invalid, default=invalid)
        assert actual is None

    valids = (
        ({"--parent-dir": str_tmp_path}, tmp_path),
        ({"parent-dir": str_tmp_path}, tmp_path),
        ({}, None),  # missing key parent-dir or --parent-dir
    )
    for d_section, expected in valids:
        actual = _parent_dir(d_section)
        assert actual == expected

    # normal usage
    contents_including_path = toml_contents.format(str_tmp_path)
    d_section = ConfigSettings.get_section_dict(tmp_path, contents_including_path)
    actual = _parent_dir(d_section)
    assert issubclass(type(actual), PurePath)
    assert actual == tmp_path


testdata_refresh = (
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/pip.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/manage.lock",
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lnk",
            "requirements/pip.lnk",
            "requirements/manage.lnk",
        ),
        True,
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        ),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/tox.in",
            "requirements/manage.in",
        ),
        (
            "requirements/prod.lock",
            "requirements/tox.lock",
            "requirements/manage.lock",
            "requirements/prod.unlock",
            "requirements/tox.unlock",
            "requirements/manage.unlock",
        ),
        (
            "requirements/prod.lnk",
            "requirements/tox.lnk",
            "requirements/manage.lnk",
        ),
        False,
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        (
            "requirements/prod.in",
            "requirements/pins.in",
            "requirements/pip.in",
            "requirements/pip-tools.in",
            "requirements/dev.in",
            "requirements/tox.in",
            "requirements/manage.in",
            "docs/requirements.in",
        ),
        (
            "requirements/prod.lock",
            "requirements/pip.lock",
            "requirements/pip-tools.lock",
            "requirements/dev.lock",
            "requirements/tox.lock",
            "requirements/manage.lock",
            "docs/requirements.lock",
            "requirements/prod.unlock",
            "requirements/pip.unlock",
            "requirements/pip-tools.unlock",
            "requirements/dev.unlock",
            "requirements/tox.unlock",
            "requirements/manage.unlock",
            "docs/requirements.unlock",
        ),
        (
            "requirements/prod.lnk",
            "requirements/pip.lnk",
            "requirements/pip-tools.lnk",
            "requirements/dev.lnk",
            "requirements/tox.lnk",
            "requirements/manage.lnk",
            "docs/requirements.lnk",
        ),
        False,
    ),
)
ids_refresh = (
    "missing tox (MissingRequirementsFoldersFiles)",
    "manage and tox",
    "constraint path needs to be resolved",
)


@pytest.mark.parametrize(
    "path_pyproject_toml, seq_create_in_files, seq_create_lock_files, seq_expected, str_returned",
    testdata_refresh,
    ids=ids_refresh,
)
def test_plugin_refresh_links_normal(
    path_pyproject_toml,
    seq_create_in_files,
    seq_create_lock_files,
    seq_expected,
    str_returned,
    tmp_path,
    prepare_folders_files,
    prep_pyproject_toml,
    path_project_base,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_plugin_refresh_links_normal" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    path_config = prep_pyproject_toml(path_pyproject_toml, tmp_path)

    # prepare
    assert isinstance(seq_create_in_files, Sequence)
    assert isinstance(seq_create_lock_files, Sequence)
    #    makes both folders and files, but blank files
    prepare_folders_files(seq_create_in_files, tmp_path)
    prepare_folders_files(seq_create_lock_files, tmp_path)

    #    move real files, no need to create folders
    path_cwd = path_project_base()
    for p_f in seq_create_in_files:
        abspath_src = path_cwd.joinpath(p_f)
        abspath_dest = tmp_path.joinpath(p_f)
        shutil.copy2(abspath_src, abspath_dest)
        # abspath_files.append(abspath_dest)
        pass

    for p_f in seq_create_lock_files:
        abspath_src = path_cwd.joinpath(p_f)
        abspath_dest = tmp_path.joinpath(p_f)
        shutil.copy2(abspath_src, abspath_dest)
        # abspath_files.append(abspath_dest)
        pass

    configs = (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
        ),
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="1"\n\n"""
        ),
    )
    for toml_contents in configs:
        d_section = ConfigSettings.get_section_dict(tmp_path, toml_contents)

        inst = BackendType.load_factory(
            path_config,
            parent_dir=tmp_path,
        )
        assert inst.parent_dir == tmp_path

        """
        with patch(
            "drain_swamp.monkey.plugins.ds_refresh_links.Path.cwd",
            return_value=tmp_path,
        ):
        """
        with (
            patch(
                f"{g_app_name}.monkey.plugins.ds_refresh_links.BackendType.load_factory",
                return_value=inst,
            ),
            patch(
                "drain_swamp.monkey.plugins.ds_refresh_links.Path.cwd",
                return_value=tmp_path,
            ),
        ):
            str_msg = before_version_infer(d_section)
            if str_returned:
                # returns error message
                assert str_msg is not None
                assert isinstance(str_msg, str)
                logger.info(str_msg)
            else:
                # success
                assert str_msg is None
                # verify symlinks -- created, w/o resolve
                for relpath_expected in seq_expected:
                    abspath_expected = tmp_path.joinpath(relpath_expected)
                    assert abspath_expected.exists() and abspath_expected.is_symlink()
                    # clean up symlink
                    abspath_expected.unlink()
                    assert not abspath_expected.exists()


def test_plugin_refresh_links_exceptions(
    tmp_path,
    prep_pyproject_toml,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_plugin_refresh_links_exceptions" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level
    configs = (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
            f"""parent-dir="{str(tmp_path)}"\n"""
        ),
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="1"\n\n"""
            f"""parent-dir="{str(tmp_path)}"\n"""
        ),
    )

    path_config = None
    for toml_contents in configs:
        d_section = ConfigSettings.get_section_dict(tmp_path, toml_contents)
        if path_config is not None and path_config.exists():
            path_config.unlink()

        with patch(
            "drain_swamp.monkey.plugins.ds_refresh_links.Path.cwd",
            return_value=tmp_path,
        ):
            # No pyproject.toml --> PyProjectTOMLReadError msg
            str_msg = before_version_infer(d_section)
            assert str_msg is not None

            if path_config is not None and path_config.exists():
                path_config.unlink()

            # Malformed pyproject.toml
            path_pyproject_toml = Path(__file__).parent.joinpath(
                "_bad_files",
                "backend_only.pyproject_toml",
            )
            path_config = prep_pyproject_toml(path_pyproject_toml, tmp_path)
            str_msg = before_version_infer(d_section)
            assert str_msg is not None
            if path_config is not None and path_config.exists():
                path_config.unlink()

    pyproject_tomls = (
        (
            Path(__file__).parent.joinpath(
                "_good_files",
                "complete-manage-pip-prod-unlock.pyproject_toml",
            ),
            None,
        ),
    )
    for path_pyproject_toml, expected_error_msg in pyproject_tomls:
        for toml_contents in configs:
            # prepare
            d_section = ConfigSettings.get_section_dict(tmp_path, toml_contents)
            path_config = prep_pyproject_toml(path_pyproject_toml, tmp_path)
            assert path_config.exists() and path_config.is_file()
            with patch(
                "drain_swamp.monkey.plugins.ds_refresh_links.Path.cwd",
                return_value=tmp_path,
            ):
                try:
                    d_pyproject_toml, path_f = BackendType.read(tmp_path)
                except (PyProjectTOMLParseError, PyProjectTOMLReadError):
                    d_pyproject_toml = None
                assert d_pyproject_toml is not None
                assert isinstance(d_pyproject_toml, Mapping)

                """BackendNotSupportedError
                Don't replace BackendType.load_factory, into config_settings
                provide parent-dir
                """
                with (
                    patch(
                        "drain_swamp.backend_abc.BackendType.determine_backend",
                        return_value="maniac_into_squats",
                    ),
                ):
                    before_version_infer(d_section)

                # normal execution
                str_error = before_version_infer(d_section)
                if expected_error_msg is None:
                    assert str_error is None
                else:
                    assert str_error is not None and str_error.startswith(
                        expected_error_msg
                    )

            if path_config is not None and path_config.exists():
                path_config.unlink()

        # prepare
        # no set-lock in config_settings --> get set-lock from pyproject.toml --> AssertionError
        toml_contents = (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            f"""parent-dir="{str(tmp_path)}"\n"""
        )
        path_pyproject_toml = Path(__file__).parent.joinpath(
            "_bad_files",
            "static_dependencies.pyproject_toml",
        )
        expected_err = (
            "Either no pyproject.toml section, tool.setuptools.dynamic "
            "or no dependencies key"
        )
        d_section = ConfigSettings.get_section_dict(tmp_path, toml_contents)
        if path_config is not None and path_config.exists():
            path_config.unlink()

        path_config = prep_pyproject_toml(path_pyproject_toml, tmp_path)
        assert path_config.exists() and path_config.is_file()
        with patch(
            "drain_swamp.monkey.plugins.ds_refresh_links.Path.cwd",
            return_value=tmp_path,
        ):
            str_error = before_version_infer(d_section)
            assert has_logging_occurred(caplog)
            assert str_error == expected_err



import traceback
import unittest
from pathlib import Path

import click
from click.testing import CliRunner

help_path = "Provide a Path. relative or absolute"

EPILOG_CHOKE = """
Description:

click.Path(resolve_path=True) assumes can resolve the user supplied input.
A float is not a Path, so cannot be resolved. Results in a TypeError and the
entrypoint crashing with a traceback.

End users should never be presented with a traceback. When click is
dealing with these issues the normal exit code is 2

Reproduce issue:

python -m unittest integration.test_upstream_click_issue

Expected behavior:

Do not make assumptions regarding the user input type. Handler should
accept typing.Any

- Ideally and gracefully, use the default as a fallback. No exit code

- Otherwise exit code 2

Why'd this happen?:

Python typing encourages parameters ideally to accept typing.Any
Unless already handled elsewhere, assuming parameters have particular
type(s) is a recipe for disaster. Should always check

Environment:

click: 8.1.7
Python: py39
OS: GNU/Linux
distro: Void Linux

Output:

  File ".venv/lib/python3.9/site-packages/click/testing.py", line 408, in invoke
    return_value = cli.main(args=args or (), prog_name=prog_name, **extra)
  File ".venv/lib/python3.9/site-packages/click/core.py", line 1077, in main
    with self.make_context(prog_name, args, **extra) as ctx:
  File ".venv/lib/python3.9/site-packages/click/core.py", line 943, in make_context
    self.parse_args(ctx, args)
  File ".venv/lib/python3.9/site-packages/click/core.py", line 1408, in parse_args
    value, args = param.handle_parse_result(ctx, opts, args)
  File ".venv/lib/python3.9/site-packages/click/core.py", line 2400, in handle_parse_result
    value = self.process_value(ctx, value)
  File ".venv/lib/python3.9/site-packages/click/core.py", line 2356, in process_value
    value = self.type_cast_value(ctx, value)
  File ".venv/lib/python3.9/site-packages/click/core.py", line 2344, in type_cast_value
    return convert(value)
  File ".venv/lib/python3.9/site-packages/click/core.py", line 2316, in convert
    return self.type(value, param=self, ctx=ctx)
  File ".venv/lib/python3.9/site-packages/click/types.py", line 83, in __call__
    return self.convert(value, param, ctx)
  File ".venv/lib/python3.9/site-packages/click/types.py", line 869, in convert
    rv = os.fsdecode(pathlib.Path(rv).resolve())
  File "~/.pyenv/versions/3.9.16/lib/python3.9/pathlib.py", line 1082, in __new__
    self = cls._from_parts(args, init=False)
  File "~/.pyenv/versions/3.9.16/lib/python3.9/pathlib.py", line 707, in _from_parts
    drv, root, parts = self._parse_args(args)
  File "~/.pyenv/versions/3.9.16/lib/python3.9/pathlib.py", line 691, in _parse_args
    a = os.fspath(a)
F
======================================================================
FAIL: test_choke_on_this (integration.test_upstream_click_issue.TestChokeOnNonPath)
----------------------------------------------------------------------
Traceback (most recent call last):
  File "integration/test_upstream_click_issue.py", line 86, in test_choke_on_this
    assert False
    cmd = ('--path', 0.1234)
    exit_code = 1
    result = <Result TypeError('expected str, bytes or os.PathLike object, not float')>
    runner = <click.testing.CliRunner object at 0x7fdb2060c4c0>
    self = <integration.test_upstream_click_issue.TestChokeOnNonPath testMethod=test_choke_on_this>
    str_tb = ...
    tb = <traceback object at 0x7fdb20b753c0>
AssertionError

----------------------------------------------------------------------
Ran 1 test in 0.003s

FAILED (failures=1)


"""


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.version_option(version="0.0.1")
def main():
    """Command-line. Prints usage"""


@main.command(
    "choke",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_CHOKE,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
def choke_on_non_path(path):
    # When a relative path, gets resolved, but returns as str rather than Path
    """
    $> cd integrations
    $> printf 'test_upstream_click_issue.py choke --path=%f' 0.1234 | xargs python
    path: /home/faulkmore/Downloads/git_decimals/drain_swamp/integration/0.123400 <class 'str'>
    """
    print(f"path: {path} {type(path)}")
    if isinstance(path, str):
        path = Path(path)


class TestChokeOnNonPath(unittest.TestCase):
    def test_choke_on_this(self):
        """This is bypassing the shell, which was the mistake. The shell
        treats everything as a string

        .. seealso::

           `click#2742 <https://github.com/pallets/click/issues/2742>`_

            Thanks `David Lord <https://github.com/davidism>`_ for your time

        """
        runner = CliRunner()
        cmd = (
            "--path",
            0.1234,  # <-- Shell normally would treat as str, not float
        )
        # result = <Result TypeError('expected str, bytes or os.PathLike object, not float')>
        result = runner.invoke(choke_on_non_path, cmd)

        # An uncatchable (within the click command, choke_on_non_path) TypeError
        # is equivalent to click crashing
        assert result.exception.__class__ == TypeError

        tb = result.exc_info[2]
        # str_tb = traceback.format_tb(tb)
        traceback.print_tb(tb)

        exit_code = result.exit_code
        # ideally use the default, Path.cwd() and exit code == 0
        assert exit_code != 2  # expecting 2
        assert exit_code == 1  # meaningless exit code

        # uncomment to see the locals
        # assert False
        pass


if __name__ == "__main__":  # pragma: no cover
    """
    .. code-block:: shell

       python -m unittest integration.test_upstream_click_issue --locals

    """
    # main()
    unittest.main(tb_locals=True)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_plugin_scm_version.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_plugin_scm_version.py

"""

import subprocess
from pathlib import Path
from unittest.mock import patch

import pytest

from drain_swamp.monkey.config_settings import ConfigSettings
from drain_swamp.monkey.plugins.ds_scm_version import (
    _is_package_installed,
    _kind,
    on_version_infer,
)

from .wd_wrapper import WorkDir


@pytest.fixture()
def wd(wd: WorkDir) -> WorkDir:
    # Create project base folder with project name
    path_new = wd.cwd / "complete_awesome_perfect"
    path_new.mkdir()
    wd.cwd = path_new

    wd("git init")
    wd("git config user.email test@example.com")
    wd('git config user.name "a test"')
    wd.add_command = "git add ."
    wd.commit_command = "git commit --no-verify --no-gpg-sign -m test-{reason}"
    return wd


testdata_is_package_installed = (
    (
        "drain_swamp",
        True,
    ),
    (
        "asdfsadfsadfsadfasdfasdfsadfsaddsafdsafdsafdsafdsafdsaf",
        False,
    ),
)
ids_is_package_installed = (
    "Installed package",
    "not installed package",
)


@pytest.mark.parametrize(
    "pkg_name, is_installed_expected",
    testdata_is_package_installed,
    ids=ids_is_package_installed,
)
def test_is_package_installed(pkg_name, is_installed_expected):
    # pytest --showlocals --log-level INFO -k "test_is_package_installed" tests
    actual = _is_package_installed(pkg_name)
    assert actual == is_installed_expected


testdata_kind_arg_filter_normal = (
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
        ),
        "tag",
        "current",
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
        ),
        None,
        "current",
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
        ),
        "tag",
        "tag",
    ),
)
ids_kind_arg_filter_normal = (
    "kind current fallback tag",
    "kind current fallback None",
    "no kind",
)


@pytest.mark.parametrize(
    "toml_contents, fallback, expected_kind",
    testdata_kind_arg_filter_normal,
    ids=ids_kind_arg_filter_normal,
)
def test_kind_arg_filter_normal(toml_contents, fallback, expected_kind, tmp_path):
    # pytest --showlocals --log-level INFO -k "test_kind_arg_filter_normal" tests
    # d_section not a dict
    invalids = (
        None,
        1.2,
    )
    for invalid in invalids:
        fallback_clean = fallback if fallback is not None else "tag"
        actual_kind = _kind(invalid, fallback=fallback_clean)
        assert actual_kind == fallback_clean

    d_section = ConfigSettings.get_section_dict(tmp_path, toml_contents)
    actual_kind = _kind(d_section, fallback=fallback)
    assert actual_kind == expected_kind

    # not used thru :code:`python -m build`
    d_section = {"--kind": "0.0.1"}
    actual_kind = _kind(d_section, fallback=fallback)
    assert actual_kind == "0.0.1"


testdata_kind_arg_filter_whitespace = (
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind="   "\n"""
            """set-lock="0"\n\n"""
        ),
        "nonsense",
        "nonsense",
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.config-settings]\n"
            """kind=""\n"""
            """set-lock="0"\n\n"""
        ),
        "nonsense",
        "nonsense",
    ),
)
ids_kind_arg_filter_whitespace = (
    "kind is just whitespace --> fallback",
    "kind is empty --> fallback",
)


@pytest.mark.parametrize(
    "toml_contents, fallback, expected_kind",
    testdata_kind_arg_filter_whitespace,
    ids=ids_kind_arg_filter_whitespace,
)
def test_kind_arg_filter_whitespace(toml_contents, fallback, expected_kind, tmp_path):
    # pytest --showlocals --log-level INFO -k "test_kind_arg_filter_whitespace" tests
    d_section = ConfigSettings.get_section_dict(tmp_path, toml_contents)
    with pytest.warns(UserWarning) as record:
        actual_kind = _kind(d_section, fallback=fallback)
        warn_msg_expected = (
            "-C--kind expected to be: tag, current, now, or a version str. Got"
        )
        warn_msg = str(record[0].message)
        assert isinstance(warn_msg, str)
        assert warn_msg.startswith(warn_msg_expected)
        assert actual_kind == expected_kind


def test_on_version_infer(
    wd: WorkDir,
    prep_pyproject_toml,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_on_version_infer" tests
    # tag. Does not modify the version file
    kind = "tag"
    d_config_settings = {"kind": "tag"}
    actual = on_version_infer(d_config_settings)
    assert actual is None

    # if drain swamp not installed, could only test against drain_swamp package
    pass

    # current now and version str acts modifies a package, so act against a fake package
    # test_wrap_infer_version.py creates a fake package environment

    # prepare
    app_name = "complete_awesome_perfect"

    #    .gitignore
    msg_gitignore = (
        "dist/*.whl\n"
        "dist/*.tar.gz\n"
        "requirements/*.lnk\n"
        "src/*.egg-info/*\n"
        "src/*/_version.py\n"
        "setuptools-build.toml\n\n"
    )
    wd.write(".gitignore", msg_gitignore)

    #    pyproject.toml
    p_toml_file = Path(__file__).parent.joinpath(
        "_project",
        "install_minimum.pyproject_toml",
    )
    path_pyproject_toml = prep_pyproject_toml(p_toml_file, wd.cwd)

    #    empty files
    req_seq = (
        "dist/empty.txt",
        f"src/{app_name}/__init__.py",
        f"src/{app_name}/__main__.py",
        f"src/{app_name}/_version.py",
        "requirements/prod.in",
        "requirements/pip.in",
        "requirements/manage.in",
        "requirements/prod.unlock",
        "requirements/pip.unlock",
        "requirements/manage.unlock",
        "README.rst",
        "LICENSE",
    )
    prepare_folders_files(req_seq, wd.cwd)
    #   Only wanted the folder
    wd.cwd.joinpath("dist", "empty.txt").unlink()

    #    src/complete_awesome_perfect/__main__.py
    path_dir_src = Path(__file__).parent.joinpath(
        "_project",
        "src",
        app_name,
    )
    path_dir_dest = wd.cwd / "src" / app_name
    path_f = path_dir_src.joinpath("__main__.py")
    prep_pyproject_toml(path_f, path_dir_dest, rename="__main__.py")

    #    src/complete_awesome_perfect/_version.py
    str_version_file = (
        """__version__ = version = '0.2.0'\n"""
        """__version_tuple__ = version_tuple = (0, 2, 0)\n\n"""
    )
    path_f = path_dir_dest.joinpath("_version.py")
    path_f.write_text(str_version_file)

    wd.add_and_commit()
    wd("git tag --no-sign -m 'plz' 0.2.0")
    #    monkeypatch.chdir(p_package_base)

    # act
    kinds = (
        "now",
        "current",
        "0.2.1.dev4",
    )
    with patch(
        "drain_swamp._run_cmd.Path.cwd",
        return_value=wd.cwd,
    ):
        if not path_pyproject_toml.exists():
            prep_pyproject_toml(p_toml_file, wd.cwd)
            assert path_pyproject_toml.exists()

        # valids
        for kind in kinds:
            d_config_settings = {"kind": kind}
            actual_msg = on_version_infer(d_config_settings)
            # msg is none if no error occurred
            assert actual_msg is None

        # invalid semantic str. Should not be able to write
        # [venv bin path]/scm-version write "golf balls are not good eating"
        invalids = (
            "golf balls are not good eating",
            "   ",
            "",
        )
        for invalid in invalids:
            d_config_settings = {"kind": invalid}
            actual_golf_balls = on_version_infer(d_config_settings)
            assert actual_golf_balls is not None
            assert isinstance(actual_golf_balls, str)

        """error during subprocess call to :code:`scm-version get` or
        :code:`scm-version write`. Reason timeout too slow"""
        with (
            patch(
                "drain_swamp._run_cmd.subprocess.run",
                side_effect=subprocess.TimeoutExpired(
                    cmd=["the", "cmd"],
                    timeout=1,
                    output=None,
                    stderr="go faster next time",
                ),
            ),
            pytest.warns() as record,
        ):
            d_config_settings = {"kind": kind}
            actual_msg = on_version_infer(d_config_settings)
            assert actual_msg is None
            str_warn = str(record[0].message)
            assert str_warn is not None
            assert isinstance(str_warn, str)

        # kind==current. scm-version get cause error
        kind_cause_error = "current"
        code = 3
        cmd = ["scm-version", "get", "--is-write"]
        stderr = "not blank means subprocess call failed"
        with (
            patch(
                "drain_swamp._run_cmd.subprocess.run",
                side_effect=subprocess.CalledProcessError(
                    code,
                    cmd,
                    output=None,
                    stderr=stderr,
                ),
            ),
            pytest.warns() as record,
        ):
            d_config_settings = {"kind": kind_cause_error}
            actual_msg = on_version_infer(d_config_settings)
            assert actual_msg is None
            str_warn = str(record[0].message)
            assert str_warn is not None
            assert isinstance(str_warn, str)

        # Cause LookupError in subprocess by removing the pyproject.toml file
        path_pyproject_toml.unlink()
        assert not path_pyproject_toml.exists()
        for kind in kinds:
            d_config_settings = {"kind": kind}
            actual_msg = on_version_infer(d_config_settings)
            assert actual_msg is not None
            assert isinstance(actual_msg, str)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. py:data:: __all__
   :type: tuple[str, str, str]
   :value: ("is_ok", "is_relative_required", "is_iterable_not_str")

   Module exports

.. py:data:: DEFAULT_EXTENSIONS
   :type: tuple[str, ...]
   :value: (".in",)

   Acceptable file extensions.

"""

from __future__ import annotations

import sys
from pathlib import PurePath

from .constants import SUFFIX_IN

if sys.version_info >= (3, 9):  # pragma: no cover
    from collections.abc import (
        Generator,
        Sequence,
    )
else:  # pragma: no cover
    from typing import (
        Generator,
        Sequence,
    )

__package__ = "drain_swamp"
__all__ = (
    "is_ok",
    "is_relative_required",
    "is_iterable_not_str",
)

DEFAULT_EXTENSIONS = (SUFFIX_IN,)
CLICK_TRUE = ("1", "true", "t", "yes", "y", "on")
CLICK_FALSE = ("0", "false", "f", "no", "n", "off")


def is_ok(test):
    """Check if non-empty str

    Edge case: contains only whitespace --> ``False``

    :param test: variable to test
    :type test: typing.Any | None
    :returns: ``True`` if non-empty str otherwise ``False``
    :rtype: bool

    """
    ret = False
    is_str = test is not None and isinstance(test, str)
    if is_str:
        # Edge case: contains only whitespace
        str_stripped = test.strip()
        ret = len(str_stripped) != 0
    else:
        ret = False

    return ret


def is_relative_required(
    path_relative=None,
    extensions=DEFAULT_EXTENSIONS,
) -> bool:
    """Check that is relative path with expected suffix

    :param path_relative: Default None. Intended to be a relative path
    :type path_relative: pathlib.Path | None
    :param extensions:

       Default :py:data:`~drain_swamp.check_type.DEFAULT_EXTENSIONS`.
       Acceptable file extensions.

    :type extensions: collections.abc.Sequence[str]
    :returns: If an acceptable relative path with appropriate suffix
    :rtype: bool
    :raises:

       - :py:exc:`ValueError` -- Expecting at least one suffix to use to compare
       - :py:exc:`TypeError` -- Unsupported type expected a sequence (tuple or list)

    """
    lst_ext = []
    if not is_iterable_not_str(extensions):
        msg_exc = "Unsupported type expected a sequence (tuple or list)"
        raise TypeError(msg_exc)
    else:  # pragma: no cover
        pass

    # If not a str, ignore it
    exts = [ext for ext in extensions if isinstance(ext, str)]

    # if needed, prepend a "."
    for ext in exts:
        if not ext.startswith("."):
            lst_ext.append(f".{ext}")
        else:
            lst_ext.append(ext)

    if len(lst_ext) == 0:
        msg_exc = (
            "Expecting at least one suffix to compare against the "
            "relative paths suffixes"
        )
        raise ValueError(msg_exc)
    else:  # pragma: no cover
        pass

    is_relative_path_acceptable = (
        path_relative is not None
        and issubclass(type(path_relative), PurePath)
        and not path_relative.is_absolute()
        and path_relative.suffixes == lst_ext
    )

    return is_relative_path_acceptable


def is_iterable_not_str(mixed):
    """Can be iterated thru, but not the false positive, str

    :param mixed: Something that hopefully can be iterated thru
    :type mixed: typing.Any | None
    :returns: True if suitable for iteration and not a scalar str otherwise False
    :rtype: bool
    """
    ret = (
        mixed is not None
        and (
            isinstance(mixed, Sequence)
            or isinstance(mixed, set)
            or isinstance(mixed, Generator)
        )
        and not isinstance(mixed, str)
    )

    return ret


def click_bool(val=None):
    """Simulate click.Bool

    :param val: str provided on the command line indicating boolean state
    :type val: str | None
    :returns:

       None if not provided. True if command line equivalent of True or
       False if the opposite equivalent

    :rtype: bool | None
    """
    if val is None:
        ret = None
    else:
        if val in CLICK_TRUE:
            ret = True
        elif val in CLICK_FALSE:
            ret = False
        else:
            ret = None

    return ret



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

``pyproject.toml`` parsing. text --> dict

.. py:data:: __all__
   :type: tuple[str]
   :value: ("TomlParser",)

   Module exports

"""

from __future__ import annotations

import sys
from pathlib import (
    Path,
    PurePath,
)
from typing import TYPE_CHECKING

from .exceptions import (
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)
from .pep518_read import find_pyproject_toml

if sys.version_info >= (3, 11):  # pragma: no cover
    try:
        import tomllib
    except ImportError:
        # Help users on older alphas
        if not TYPE_CHECKING:
            import tomli as tomllib
else:  # pragma: no cover
    import tomli as tomllib

__package__ = "drain_swamp"
__all__ = ("TomlParser",)


class TomlParser:
    """Reverse searches for a ``pyproject.toml`` file and parses it.

    Interface provides both the dict and the resolved absolute path

    :ivar path:

       Starting search absolute path. Can be a folder or file. Reverse
       searches for a ``pyproject.toml`` file

    :vartype path: typing.Any | None
    :ivar raise_exceptions: Default False. Expecting a bool. If True raise exceptions
    :vartype raise_exceptions: typing.Any | None
    :raises:

       - :py:exc:`drain_swamp.exceptions.PyProjectTOMLReadError` --
         pyproject.toml either doesn't exist or inaccessible

       - :py:exc:`drain_swamp.exceptions.PyProjectTOMLParseError` --
         pyproject.toml unparsable

    """

    def __init__(self, path, raise_exceptions=False):
        super().__init__()
        if raise_exceptions is None or not isinstance(raise_exceptions, bool):
            raise_exceptions = False
        else:  # pragma: no cover
            pass

        """
        is_dir = (
            path is not None
            and issubclass(type(path), PurePath)
            and path.exists()
            and path.is_dir()
            and path.joinpath("pyproject.toml").exists()
            and path.joinpath("pyproject.toml").is_file()
        )
        is_pyproject_toml = (
            path is not None
            and issubclass(type(path), PurePath)
            and path.exists()
            and path.is_file()
            and path.name == "pyproject.toml"
        )

        if is_dir:
            path_config = path.joinpath("pyproject.toml")
        elif is_pyproject_toml:
            path_config = path
        else:
            path_config = None
        """
        if path is None or not issubclass(type(path), PurePath):
            path_config = None
        else:
            path_config = path

        try:
            d_pyproject_toml = self._get_pyproject_toml(path_config)
        except PyProjectTOMLParseError:
            if raise_exceptions is True:
                raise
            else:
                d_pyproject_toml = None
        except (TypeError, FileNotFoundError) as e:
            if raise_exceptions is True:
                msg_warn = "pyproject.toml is either not a file or lacks r/w permission"
                raise PyProjectTOMLReadError(msg_warn) from e
            else:
                d_pyproject_toml = None

        self._d_pyproject_toml = d_pyproject_toml

    @property
    def path_file(self):
        """Absolute Path to ``pyproject.toml``

        :returns:

           Absolute Path to ``pyproject.toml`` file. Not set if constructor
           raised an Exception

        :rtype: pathlib.Path | None
        """
        return self._path_file

    @property
    def d_pyproject_toml(self):
        """Getter pyproject.toml dict

        :returns: pyproject.toml dict. Not set if constructor raised an Exception
        :rtype: collections.abc.Mapping[str, typing.Any] | None
        """
        return self._d_pyproject_toml

    @classmethod
    def resolve(cls, path_config):
        """Reverse search for a ``pyproject.toml`` file

        Leverages reverse search algo from Python package, black

        :param path_config:

           Starting absolute path to reserve search for a ``pyproject.toml``.
           Expecting an absolute path either: Path or str

        :type path_config: typing.Any | None
        :returns: absolute path to a pyproject.toml
        :rtype: pathlib.Path
        :raises:

           - :py:exc:`TypeError` -- Start search path unsupported type
           - :py:exc:`FileNotFoundError` -- Reverse search pyproject.toml not found

        """
        meth_name = "resolve_pyproject_toml"
        msg_exc_type_bad = f"Unsupported type expecting a Path. Got {type(path_config)}"
        is_type_ng = path_config is None or not (
            isinstance(path_config, str) or issubclass(type(path_config), PurePath)
        )
        if is_type_ng:
            # unsupported type
            raise TypeError(msg_exc_type_bad)
        else:  # pragma: no cover
            pass

        if isinstance(path_config, str):
            path_file = Path(path_config)
        else:
            path_file = path_config

        msg_exc_no_such_file = (
            f"In {meth_name}, positional arg, no such file {str(path_file)}"
        )

        # can be a file or dir or symlink. Try to resolve
        path_dir = path_file
        t_str = (str(path_dir),)
        file_path = find_pyproject_toml(t_str, None)
        if file_path is None:
            raise FileNotFoundError(msg_exc_no_such_file)
        else:
            path_file = Path(file_path)

        return path_file

    def _get_pyproject_toml(self, path_config):
        """Load pyproject.toml

        Package click handles converting to :py:class:`pathlib.Path`.

        Example usage

        .. code-block:: text

           from drain_swamp.parser_in import get_pyproject_toml

           path_config = Path("[proj path]/requirements/prod.in")
           d_pyproject_toml = get_pyproject_toml(path_config)
           config_tables = d_pyproject_toml.get("tool", {}).get("setuptools", {}).get("dynamic", {})
           for (key, value) in config_tables.items():
               ...

        :param path_config: absolute path to ``pyproject.toml`` file or it's folder
        :type path_config: typing.Any | None
        :returns: tomllib dict. Cannot know yet which fields are needed
        :rtype: collections.abc.Mapping[str, typing.Any]
        :raises:

           - :py:exc:`drain_swamp.exceptions.PyProjectTOMLParseError` --
             parsing toml unsuccessful

           - :py:exc:`TypeError` -- positional arg unsupported type

           - :py:exc:`FileNotFoundError` -- positional arg does not exist or not a file

        .. todo:: demand for setup.cfg / setup.py ?

           There is no support for setup.cfg / setup.py

           Should there be?

        .. todo:: check r/w permissions

           Check file is read write. If not raise a PermissionError

        """
        cls = type(self)
        # may raise TypeError or FileNotFoundError
        self._path_file = cls.resolve(path_config)

        # tomllib insists on opened as binary
        with self._path_file.open(mode="rb") as f:
            try:
                d_ret = tomllib.load(f)
            except tomllib.TOMLDecodeError as e:
                msg_exc_pyproject_toml = (
                    f"pyproject.toml either not found or cannot be parsed. "
                    f"path {self._path_file!s} exact issue: {e}"
                )
                raise PyProjectTOMLParseError(msg_exc_pyproject_toml)

        return d_ret



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

igor.py utils

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

.. py:data:: __all__
   :type: tuple[str, str, str, str, str, str]
   :value: ("seed_changelog", "edit_for_release", "build_package", \
   "pretag", "print_cheats", "get_tag_version", "write_version_file")

.. py:data:: SCRIV_START
   :type: str
   :value: ".. scriv-start-here\n\n"

   In CHANGES.rst, token beneath which a changelog entry is written

.. py:data:: UNRELEASED
   :type: str
   :value: "Unreleased\n----------"

   Written into CHANGES.rst, the start of what is being called, the seed

.. py:data:: REGEX_COPYRIGHT_LINE
   :type: str

   NOTICE.txt has a start and end year. Regex for grabbing both.

.. py:data:: COPYRIGHT_START_YEAR_FALLBACK
   :type: int
   :value: 1970

   Fallback start year. NOTICE.txt start year gets updated. Since the
   start year is static expected to be provided in pyproject.toml

"""

import os
import re
import subprocess
import sys
from pathlib import (
    Path,
    PurePath,
)

from ._run_cmd import run_cmd
from .check_type import is_ok
from .package_metadata import PackageMetadata
from .parser_in import TomlParser
from .snippet_sphinx_conf import SnipSphinxConf
from .version_file.dump_version import write_version_files
from .version_semantic import (
    SemVersion,
    _current_version,
    _path_or_cwd,
    _scm_key,
    get_package_name,
    sanitize_tag,
)

__all__ = (
    "seed_changelog",
    "edit_for_release",
    "build_package",
    "pretag",
    "print_cheats",
    "get_tag_version",
    "write_version_file",
)
__package__ = "drain_swamp"

UNRELEASED = "Unreleased\n----------"
SCRIV_START = ".. scriv-start-here\n\n"
COPYRIGHT_START_YEAR_FALLBACK = 1970
REGEX_COPYRIGHT_LINE = r"Copyright {0}.*? {1}"


def update_file(fname, pattern, replacement):
    """Update the contents of a file, replacing pattern with replacement.

    :param fname: str absolute path
    :type fname: str
    :param pattern: regex pattern
    :type pattern: str
    :param replacement: replacement text after regex substitution
    :type replacement: str

    .. todo:: check have r/w permissions

       Did not check file permissions

    """
    is_path_or_str = is_ok(fname) or issubclass(type(fname), PurePath)
    if not is_path_or_str:
        msg_exc = f"Expected a non-empty str path or Path got {fname}"
        print(msg_exc, file=sys.stderr)
        return

    try:
        path_file = Path(fname)
    except Exception:  # pragma: no cover
        path_file = None

    is_absolute_file = (
        path_file is not None and path_file.exists() and path_file.is_file()
    )
    if is_absolute_file:
        # Assumes r/w permissions
        old_text = path_file.read_text()

        new_text = re.sub(pattern, replacement, old_text, count=1)

        if new_text != old_text:
            msg_info = f"Updating {fname}"
            print(msg_info, file=sys.stderr)

            path_file.write_text(new_text)
        else:  # pragma: no cover
            # Do nothing
            pass
    else:
        msg_exc = f"Provide an absolute path to a file. Got, {fname}"
        print(msg_exc, file=sys.stderr)
        return


def seed_changelog(path_cwd):
    """Add a new empty changelog entry

    :param path_cwd: Current working directory path
    :type path_cwd: pathlib.Path
    """
    path_changelog = path_cwd.joinpath("CHANGES.rst")
    pattern = re.escape(SCRIV_START)
    replacement = f"{UNRELEASED}\n\nNothing yet.\n\n\n" + SCRIV_START
    update_file(
        path_changelog,
        pattern,
        replacement,
    )


def edit_for_release(path_cwd, kind, snippet_co=None):
    """Edit a few files in preparation for a release.

    .. code-block:: text

       [tool.drain-swamp]
       copyright_start_year = [an int]

    app name also taken from pyproject.toml

    :param path_cwd: Current working directory path
    :type path_cwd: pathlib.Path
    :param kind:

       version str, "tag", or "current" or "now". For tagged versions, a version str

    :type kind: str
    :param snippet_co:

       Default None. Sphinx conf.py snippet code

    :type snippet_co: str
    :returns:

       None if no error occurred. 3 if no folder, 4 if no doc?/conf.py
       file. 2 version str was dodgy

    :rtype: int | None
    """
    # user input parsing?
    path_changelog = path_cwd.joinpath("CHANGES.rst")
    path_notice_txt = path_cwd.joinpath("NOTICE.txt")

    try:
        sc = SnipSphinxConf(path=path_cwd)
    except NotADirectoryError:
        msg_exc = "Expected a doc/ or docs/ folder"
        print(msg_exc, file=sys.stderr)
        return 3
    except FileNotFoundError:
        # if no doc?/conf.py ... do nothing
        msg_exc = "As an optimization, doc?/conf.py is needed. It can be an empty file"
        print(msg_exc, file=sys.stderr)
        return 4

    # sc.path_cwd if already filtered
    # ensure app name is not a possible package, so pyproject.toml is loaded
    pm = PackageMetadata(UNRELEASED, path=sc.path_cwd)
    d_pyproject_toml = pm.d_pyproject_toml

    # Get copyright_start_year from pyproject.toml
    d_section = d_pyproject_toml.get("tool", {}).get("drain-swamp", {})
    if "copyright_start_year" in d_section.keys() and (
        isinstance(d_section["copyright_start_year"], int)
        or isinstance(d_section["copyright_start_year"], str)
    ):
        try:
            copyright_start_year = int(d_section["copyright_start_year"])
        except Exception:
            """alternative is to get as a command option, but that would
            be tedious to provide every time"""
            copyright_start_year = COPYRIGHT_START_YEAR_FALLBACK
    else:
        copyright_start_year = COPYRIGHT_START_YEAR_FALLBACK

    # Get app name from pyproject.toml
    project_name = d_pyproject_toml.get("project", {}).get("name", None)
    if project_name is not None and isinstance(project_name, str):
        package_name = project_name.replace("-", "_")
    else:
        package_name = None

    # After call, will have sc.SV and sc.author_name_left
    try:
        sc.contents(
            kind,
            package_name,
            copyright_start_year,
        )
    except (AssertionError, ValueError) as e:
        msg_exc = str(e)
        print(msg_exc, file=sys.stderr)
        return 2

    """No feedback if ``doc?/conf.py`` encounters:

    - ReplaceResult.VALIDATE_FAIL

    - ReplaceResult.NO_MATCH

    """
    sc.replace(snippet_co=snippet_co)
    author_first_name = sc.author_name_left

    is_dev = sc.SV.dev is not None and isinstance(sc.SV.dev, int)
    if is_dev:
        msg_info = f"**\n** This is a dev release: {sc.SV.__version__}\n**\n\nNo edits"
        print(msg_info, file=sys.stderr)
        return

    # NOTICE.txt
    regex_pattern = REGEX_COPYRIGHT_LINE.format(
        str(copyright_start_year),
        author_first_name,
    )
    str_year_end = SnipSphinxConf.now_to_str("%Y")
    str_dt_ymd = SnipSphinxConf.now_to_str("%Y-%m-%d")

    replace_with = f"Copyright {0}-{1} {author_first_name}".format(
        str(copyright_start_year),
        str_year_end,
        author_first_name,
    )
    update_file(path_notice_txt, regex_pattern, replace_with)

    # CHANGES.rst
    ver = sc.SV.__version__
    anchor = sc.SV.anchor()
    if ver is not None and anchor is not None:
        title = f"Version {ver} — {str_dt_ymd}"
        rule = "-" * len(title)
        new_head = f".. _changes_{anchor}:\n\n{title}\n{rule}"

        pairs = (
            (re.escape(SCRIV_START), ""),
            (re.escape(UNRELEASED), f"{SCRIV_START}{new_head}"),
        )

        for pattern, replacement in pairs:
            update_file(path_changelog, pattern, replacement)
        else:  # pragma: no cover
            pass
    else:  # pragma: no cover
        pass


def get_version_file_path(path):
    """Get version file relative path, from pyproject.toml,
    tool.pipenv-unlock.version_file

    :param path: absolute path to either package base folder or ``pyproject.toml``
    :type path: pathlib.Path
    :returns:

       version file relative path. Relative to package base path. None
       if issue with ``pyproject.toml``

    :rtype: str | None
    """
    path_ = _path_or_cwd(path)
    tp = TomlParser(path_)
    d_pyproject_toml = tp.d_pyproject_toml
    if d_pyproject_toml is not None:
        ret = (
            d_pyproject_toml.get("tool", {})
            .get("pipenv-unlock", {})
            .get("version_file", None)
        )
    else:
        ret = None

    return ret


class AlterEnv:
    """setuptools thin wrapped build-backend will need to modify the
    in-process environment to inform setuptools-scm which version to use.

    That will mean getting the key / value only via subprocess to an
    pep366 compliant entrypoint

    :ivar path: Current working directory path
    :vartype path: pathlib.Path
    :ivar kind:

       version str, "tag", or "current" or "now". For tagged versions, a version str

    :vartype kind: str
    :raises:

       - :py:exc:`NotADirectoryError` -- cwd is not a folder

       - :py:exc:`AssertionError` -- Could not get package name from pyproject.toml

       - :py:exc:`ValueError` -- Explicit version str invalid

    """

    __slots__ = (
        "_path_cwd",
        "_pkg_name",
        "_scm_override_val",
        "_scm_override_key",
        "_version_file",
    )

    def __init__(self, path, kind):
        super().__init__()
        try:
            sv = SemVersion(path=path)
        except (NotADirectoryError, FileNotFoundError) as e:
            msg_exc = f"Expecting a folder path. Got a file {path}.\n{str(e)}"
            raise NotADirectoryError(msg_exc) from e

        try:
            # fallback version "0.0.1"
            # afterwards path available as sv.path_cwd
            ver = sv.version_clean(kind)
        except AssertionError:
            """Has pyproject.toml configuration issues or version file is bad
            Test: kind="tag", package_name="asdfasfsadfasdfasdf"
            """
            raise
        except ValueError:
            """Explicit version str invalid. Either from version_file,
            scm (aka git), or explicit e.g. kind='golf balls'"""
            raise

        self._path_cwd = sv.path_cwd
        self._scm_override_val = ver

        pkg_name = get_package_name(path)
        if pkg_name is None:
            msg_exc = "Could not get package name from pyproject.toml"
            raise AssertionError(msg_exc)
        else:  # pragma: no cover
            pass
        self._pkg_name = pkg_name

        self.scm_key = _scm_key(pkg_name)

        # TypeError or ValueError --> AssertionError
        version_file = get_version_file_path(path)
        try:
            self.version_file = version_file
        except TypeError as e:
            msg_warn = "Expecting a relative path to version file"
            raise AssertionError(msg_warn) from e

    def modify_env(self):
        """Get a modified :py:data:`os.environ`
        Alter environment so setuptools-scm knows what the version should be

        :returns: Modified os.environ Pass to a subprocess
        :rtype: os.Environ[typing.Any]
        """
        env = os.environ.copy()
        env |= {self.scm_key: self.scm_val}

        return env

    @property
    def path_cwd(self):
        """current working directory path.

        SemVersion does a good job at user input validation

        :returns: cwd path
        :rtype: pathlib.Path
        """
        return self._path_cwd

    @property
    def scm_key(self):
        """setuptools-scm key

        :returns: setuptools-scm key for forcing a particular version
        :rtype: str
        """
        return self._scm_override_key

    @scm_key.setter
    def scm_key(self, val):
        """Setter for scm_key. Ensure upper case and underscore, not hyphen

        :param val: Value to be set
        :type val: typing.Any
        """
        if val is not None and isinstance(val, str):
            val = val.upper()
            ret = val.replace("-", "_")
            self._scm_override_key = ret
        else:  # pragma: no cover
            pass

    @property
    def scm_val(self):
        """setuptools-scm value

        :returns: setuptools-scm val for forcing a particular version
        :rtype: str
        """
        return self._scm_override_val

    @property
    def pkg_name(self):
        """Package name from pyproject.toml project.name

        :returns: Package name
        :rtype: str
        """
        return self._pkg_name

    @property
    def version_file(self):
        """version file relative path

        :returns: version file relative path
        :rtype: str | None
        """
        return self._version_file

    @version_file.setter
    def version_file(self, val) -> None:
        """Version file setter
        :param val: relative Path to version file
        :type val: typing.Any | None
        :raises:

           - :py:exc:`TypeError` -- expecting a Path

        """
        if val is None or not isinstance(val, str):
            msg_exc = "Could not get version_file from pyproject.toml"
            raise TypeError(msg_exc)
        else:
            self._version_file = val


def get_tag_version(path):
    """Specifically want version from version file, not current version
    from setuptools-scm and git.

    If ``pyproject.toml`` is missing or malformed raise :py:exc:`AssertionError`.

    For all other issues, issue a warning and fallback to current version.
    Always want a semantic version str, even if it's wrong. The warning will indicate
    what needs to be fixed

    :param path: current working directory path
    :type path: pathlib.Path
    :param is_test: Default False. During testing set to True
    :type is_test: bool | None
    :returns: tag version fall back to current version
    :rtype: str
    :raises:

       - :py:exc:`NotADirectoryError` -- cwd is not a folder

       - :py:exc:`AssertionError` -- pyproject.toml is missing or malformed

       - :py:exc:`ValueError` -- Explicit version str invalid

    """
    kind = "tag"
    try:
        ae = AlterEnv(path, kind)
    except (ValueError, AssertionError):
        raise

    # tag version (from version file) fallback current version
    # catpure warnings and redirect to stderr
    ver = ae.scm_val

    return ver


def write_version_file(path, kind, is_test=False):
    """Writes version file. Uses vendored setuptools-scm code

    From pyproject.toml requires:

    - project.name

    - tool.pipenv-unlock.version_file

    :param path: current working directory path
    :type path: pathlib.Path
    :param kind:

       version str, "tag", or "current" or "now". For tagged versions, a version str

    :type kind: str
    :param is_test: Default False. During testing set to True
    :type is_test: bool | None
    :raises:

       - :py:exc:`NotADirectoryError` -- cwd is not a folder

       - :py:exc:`AssertionError` -- Could not get package name from pyproject.toml

       - :py:exc:`ValueError` -- Explicit version str invalid

       - :py:exc:`ValueError` -- Only support write to .py or .txt file

    """
    if is_test is None or not isinstance(is_test, bool):
        is_test = False
    else:  # pragma: no cover
        pass

    try:
        ae = AlterEnv(path, kind)
    except (ValueError, AssertionError):
        raise

    ver = ae.scm_val
    root = ae.path_cwd
    if is_test:
        write_to = ae.version_file
        version_file = None
    else:  # pragma: no cover
        write_to = None
        version_file = ae.version_file

    try:
        write_version_files(ver, root, write_to, version_file)
    except ValueError:
        # if is_test=True does not validate the template, so ValueError
        # target not a ``.txt`` or ``.py`` file
        raise


def build_package(path, kind):
    """Build package

    Replaces / obsoletes

    .. code-block:: shell

       python igor.py build_next "0.0.1"
       python igor.py build_next "tag"
       python igor.py build_next "current"
       python igor.py build_next "now"

    :param path: Current working directory path
    :type path: pathlib.Path
    :param kind:

       version str, "tag", or "current" or "now". For tagged versions, a version str

    :type kind: str
    :returns: True if build succeeded otherwise False
    :rtype: bool
    :raises:

       - :py:exc:`NotADirectoryError` -- cwd is not a folder

       - :py:exc:`AssertionError` -- Could not get package name from git

       - :py:exc:`ValueError` -- Explicit version str invalid

    .. caution:: side effect

       kind will update [app name]/src/_version.py (done by setupttools-scm)

    """
    # expecting Path cwd
    try:
        ae = AlterEnv(path, kind)
    except Exception:
        raise
    env = ae.modify_env()

    # get lock state --> send to backend
    # 0 to lock; 1 to unlock
    # --config-setting="--set-lock=1"
    cmd = [sys.executable, "-m", "build"]
    t_out = run_cmd(cmd, cwd=ae.path_cwd, env=env)
    out, err, exit_code, str_exception = t_out
    if str_exception is not None:
        """setuptools-scm requires at least one commit. Could not get
        semantic version"""
        print(str_exception, file=sys.stderr)
        ret = False
    else:  # pragma: no cover
        is_fail = exit_code != 0
        if is_fail:
            print(out, file=sys.stderr)
            ret = False
        else:
            ret = True

    return ret


def pretag(tag):
    """Idiot check / sanitize proposed tag

    Validates semantic version str. Printing the fix semantic tag.

    :param tag:

       A semantic version str. Not current tag or now.

    :type tag: str
    :returns: False if an error occurred otherwise True
    :rtype: tuple[bool, str]
    """
    try:
        clean_tag, local = sanitize_tag(tag)
    except ValueError as e:
        msg = str(e)
        bol_ret = False
    else:
        msg = clean_tag
        bol_ret = True

    return bol_ret, msg


def get_current_version(path):
    """Get current version from setuptools-scm

    :param path: current working directory path
    :type path: pathlib.Path
    :returns:

       package current (semantic) version. None if
       :py:mod:`setuptools-scm <setuptools_scm>` is not installed

    :rtype: str | None
    """
    ret = _current_version(path=path)

    return ret


def _get_branch() -> str:
    """From git, get the current branch name

    :returns: Branch name
    :rtype: str
    :meta private:
    """
    ret = subprocess.getoutput("git rev-parse --abbrev-ref @")
    return ret


def print_cheats(path, kind):
    """Print cheats

    :param path: current working directory path
    :type path: pathlib.Path
    :param kind:

       version str, "tag", or "current" or "now". For tagged versions, a version str

    :type kind: str
    :raises:

       - :py:exc:`NotADirectoryError` -- cwd is not a folder

       - :py:exc:`AssertionError` -- Could not get package name from git

       - :py:exc:`ValueError` -- Explicit version str invalid

    """
    REPO_URL = ""
    t_repo_key = ("Source code", "Repository")

    # Initial version these will both fail
    branch = _get_branch()
    sha = subprocess.getoutput("git rev-parse @")

    # expecting Path cwd
    try:
        sv = SemVersion(path=path)
    except (NotADirectoryError, FileNotFoundError) as e:
        raise NotADirectoryError(str(e)) from e

    try:
        # afterwards path available as sv.path_cwd
        ver = sv.version_clean(kind)
    except AssertionError:
        """Could not get package name from git.
        Test: kind="tag", package_name="asdfasfsadfasdfasdf"
        """
        raise
    except ValueError:
        """Explicit version str invalid.
        Test: kind='dog food tastes better than this'"""
        raise

    sv.parse_ver(ver)
    anchor = sv.anchor()

    tp = TomlParser(path)
    d_pyproject_toml = tp.d_pyproject_toml
    if d_pyproject_toml:
        # PROJECT_NAME
        PROJECT_NAME = d_pyproject_toml.get("project", {}).get("name", None)
        # REPO_URL
        d_project_urls = d_pyproject_toml.get("project", {}).get("urls", {})
        for k, v in d_project_urls.items():
            is_repo_key = REPO_URL == "" and k in t_repo_key
            if is_repo_key and isinstance(v, str) and len(v.strip()) != 0:
                REPO_URL = v
    else:  # pragma: no cover
        # If no pyproject.toml or project.name is not set, just abort
        PROJECT_NAME = ""

    is_pyproject_toml = PROJECT_NAME is not None
    if is_pyproject_toml:
        prints = []

        intro_ = f"{PROJECT_NAME} version is {ver}"
        prints.append(intro_)

        egg = f"egg={PROJECT_NAME}==0.0"  # to force a re-install

        rtd_uri = f"https://{PROJECT_NAME}.readthedocs.io"
        changes_page = f"{rtd_uri}/en/{ver}/changes.html#changes-{anchor}"
        prints.append(changes_page)

        gh_commenting = (
            "\n## For GitHub commenting:\n"
            "This is now released as part of "
            f"[{PROJECT_NAME} {ver}]("
            f"https://pypi.org/project/{PROJECT_NAME}/{ver})."
        )
        prints.append(gh_commenting)

        section_run = "\n## To run this code:"
        prints.append(section_run)

        if branch in ("master", "main"):
            msg = f"python3 -m pip install git+{REPO_URL}#{egg}"
            prints.append(msg)
        else:
            msg = f"python3 -m pip install git+{REPO_URL}@{branch}#{egg}"
            prints.append(msg)

        msg = f"python3 -m pip install git+{REPO_URL}@{sha[:20]}#{egg}"
        prints.append(msg)

        msg = (
            "\n## For other collaborators:\n"
            f"git clone {REPO_URL}\n"
            f"cd {PROJECT_NAME}\n"
            f"git checkout {sha}"
        )
        prints.append(msg)

        str_print = "\n".join(prints)
        print(str_print)
    else:  # pragma: no cover
        pass
